# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import argparse
from pathlib import Path

from ms_service_profiler.exporters.utils import check_input_path_valid, check_output_path_valid


def add_summary_exporter(func):
    from msserviceprofiler.ms_service_profiler_ext.exporters.exporter_summary import ExporterSummary

    def wrapper(args):
        default_exporters = func(args)

        summary_exporter = ExporterSummary()
        summary_exporter.initialize(args)

        return [summary_exporter] + default_exporters
    return wrapper


def arg_parse(subparsers):
    parser = subparsers.add_parser(
        "analyze", formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="MS Server Profiler Analyze"
    )
    parser.add_argument(
        '--input-path', required=True, type=check_input_path_valid, help='Path to the folder containing profile data.'
    )
    parser.add_argument(
        '--output-path',
        type=check_output_path_valid,
        default=os.path.join(os.getcwd(), 'output'),
        help='Output file path to save results.')
    parser.add_argument(
        '--log-level',
        type=str,
        default='info',
        choices=['debug', 'info', 'warning', 'error', 'fatal', 'critical'],
        help='Log level to print.')
    parser.add_argument(
        '--format', nargs='+', default=['json', 'csv', 'db'], choices=['json', 'csv', 'db'], help='Format to save.'
    )
    parser.set_defaults(func=main)


def main(args):
    from ms_service_profiler.parse import parse, preprocess_prof_folders
    from ms_service_profiler.plugins import custom_plugins
    from ms_service_profiler.utils.log import set_log_level
    from ms_service_profiler.exporters.factory import ExporterFactory
    from ms_service_profiler.exporters.utils import create_sqlite_db

    # 初始化日志等级
    set_log_level(args.log_level)

    # msprof预处理
    preprocess_prof_folders(args.input_path)

    # 初始化Exporter
    wrapped_create_exporters = add_summary_exporter(ExporterFactory.create_exporters)
    exporters = wrapped_create_exporters(args)

    # 创建output目录
    Path(args.output_path).mkdir(parents=True, exist_ok=True)
    create_sqlite_db(args.output_path)

    # 解析数据并导出
    parse(args.input_path, custom_plugins, exporters, args=args)



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os
import re
from loguru import logger

import pandas as pd


def find_first_simulate_csv(input_path_2):
    # 检查输入路径是否为有效目录
    if not os.path.exists(input_path_2) or not os.path.isdir(input_path_2):
        raise NotADirectoryError("The provided path is not a valid directory.")

    # 构建查找模式
    pattern = os.path.join(input_path_2, "simulate*.csv")

    # 查找所有匹配的文件
    files = glob.glob(pattern)

    # 检查是否有匹配的文件
    if not files:
        raise FileNotFoundError("No CSV files starting with 'simulate' found in the directory.")

    # 按文件名排序
    files.sort()

    # 返回第一个文件的路径
    return files[0]


def analyze(input_path_1, input_path_2):
    profiling_path = os.path.join(input_path_1, 'request.csv')
    df3 = pd.read_csv(profiling_path, header=0)

    total_req = df3.shape[0]
    filtered_df = df3[df3['reply_token_size'].notna()]
    success_req = filtered_df.shape[0]

    batch_path = os.path.join(input_path_1, 'batch.csv')
    df1 = pd.read_csv(batch_path, header=0)
    df1 = df1[df1['name'] == 'modelExec']
    simulate_path = find_first_simulate_csv(input_path_2)
    df2 = pd.read_csv(simulate_path, header=None)
    column_name = ['simulate_time']
    df2.columns = column_name
    # 确认两文件的行数相同
    if len(df1) != len(df2):
        raise ValueError("两个CSV文件的行数必须相同")

    # 将第二个CSV文件中的`during_time`列添加到第一个CSV文件中，并修改列名
    df1['simulate_time'] = df2['simulate_time'] / 10 ** 6

    total_simulate_time = 0
    total_decode_simulate_time = 0
    df3.sort_values(by='http_rid', ascending=True, inplace=True)
    df_prefill = df1[df1['batch_type'] == 'prefill']
    for _, row in df_prefill.iterrows():

        digits = re.findall(r'\d+', row['reqinfo'])

        # 将这些字符串转换为整数列表
        reqinfo_list = [int(num) for num in digits]

        # 使用列表推导式筛选偶数位的数字
        non_zero_values = [x for i, x in enumerate(reqinfo_list) if i % 2 == 0]
        # 遍历这些值
        for val in non_zero_values:

            if pd.isna(df3.iloc[val]['reply_token_size']):
                continue
            during_time = df3.iloc[val]['first_token_latency']
            decode_time = (df3.iloc[val]['execution_time(microsecond)'] - df3.iloc[val]['first_token_latency'])
            arrive_time = row['start_time(microsecond)'] - during_time
            complete_time = row['start_time(microsecond)'] + decode_time
            filtered_df1 = df1[df1['start_time(microsecond)'] > arrive_time]
            filtered_df2 = df1[df1['start_time(microsecond)'] < complete_time]
            # 进一步筛选出 start_time 小于等于 row['start_time'] 的行
            filtered_df1 = filtered_df1[filtered_df1['start_time(microsecond)'] <= row['start_time(microsecond)']]
            filtered_df2 = filtered_df2[filtered_df2['start_time(microsecond)'] >= row['start_time(microsecond)']]
            # 计算 simulate_time 的总和
            total_simulate_time += filtered_df1['simulate_time'].sum()
            total_decode_simulate_time += filtered_df2['simulate_time'].sum()

    df3['completed_time'] = df3['start_time_httpReq(microsecond)'] + df3['execution_time(microsecond)']
    total_latency = df3['first_token_latency'].sum() + total_simulate_time
    total_token = df3['reply_token_size'].sum()
    avg_prefill_latency = total_latency / success_req / 10 ** 6
    total_time = df3['completed_time'].max() - df3['start_time_httpReq(microsecond)'].min() + df1['simulate_time'].sum()
    throughput = total_token / total_time * 10 ** 6
    total_decode_time = df3['execution_time(microsecond)'].sum() + total_decode_simulate_time - df3[
        'first_token_latency'].sum()
    average_decode_latency = total_decode_time / (total_token - success_req) / 10 ** 6
    success_precent = success_req / total_req
    return throughput, avg_prefill_latency, average_decode_latency, success_precent





# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import threading

import yaml


CONFIG_FILE = os.path.join(os.path.dirname(__file__), 'secur_config.yaml')

_config_lock = threading.Lock()
_config_local = threading.local()


def load_config():
    global _config_local

    with _config_lock:
        if not hasattr(_config_local, 'config'):
            if not os.path.exists(CONFIG_FILE):
                raise FileNotFoundError(f"Configuration file {CONFIG_FILE} not found.")
            with open(CONFIG_FILE, 'r') as f:
                try:
                    config = yaml.safe_load(f)
                except yaml.YAMLError as e:
                    raise ValueError(f"Invalid YAML format in {CONFIG_FILE}: {e}") from e
            _config_local.config = config
            

    return _config_local.config


def get_file_size_config():
    config = load_config()

    if "PathConfig" not in config:
        raise ValueError("Cannot find PathConfig in the secur_config.yaml")

    path_config = config['PathConfig']

    if "FileSizeConfig" not in path_config:
        raise ValueError("Cannot find FileSizeConfig in the secur_config.yaml")
    
    file_size_config = path_config['FileSizeConfig']

    if "ext_mapping" not in file_size_config:
        raise ValueError("Cannot find ext_mapping in the secur_config.yaml")

    ext_mapping = file_size_config['ext_mapping']

    if not isinstance(ext_mapping, dict):
        raise ValueError("ext_mapping is not defined as a dict in the secur_config.yaml")

    file_size_config.setdefault("require_confirm", False)

    return file_size_config



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sqlite3
import argparse
import shutil
from pathlib import Path
from contextlib import contextmanager

import pandas as pd

from ms_service_profiler.exporters.utils import check_output_path_valid
from msserviceprofiler.ms_service_profiler_ext.common.sec import list_dir_common_check


@contextmanager
def connect_db(db_path):
    connection = None    
    try:
        connection = sqlite3.connect(db_path)
        yield connection
        connection.commit()
    except:
        if connection:
            connection.rollback()
        raise
    finally:
        if connection:
            connection.close()


def process_file_pairs(file_pairs, db_conn, excel_writer):
    from ms_service_profiler.utils.log import logger
    from msserviceprofiler.ms_service_profiler_ext.compare_tools import CSVComparator, DBComparator

    comparators = {'.csv': CSVComparator, '.db': DBComparator}

    for file_a, file_b in file_pairs:
        ext = Path(file_a).suffix.lower()
        logger.info("Begin to compare %r and %r", file_a, file_b)
        for comparator_cls in comparators.values():
            if not comparator_cls.supports(ext):
                continue
            comparator = comparator_cls(db_conn, excel_writer)
            try:
                comparator.process(file_a, file_b)
            except Exception as e:
                logger.error(
                    "During comparing %r and %r, there is an error ocurred: %r", file_a, file_b, e
                )
            logger.info("End to compare %r and %r", file_a, file_b)


def process_files(file_pairs, output_db, output_excel):
    with connect_db(output_db) as db_conn:
        with pd.ExcelWriter(output_excel, engine='openpyxl') as excel_writer:
            process_file_pairs(file_pairs, db_conn, excel_writer)

    shutil.copy(
        Path(__file__).parent / 'compare_tools' / 'compare_visualization.json',
        Path(output_db).with_name("compare_visualization.json")
    )


def arg_parse(subparsers):
    parser = subparsers.add_parser(
        "compare", formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="MS Server Profiler Compare Tool"
    )
    parser.add_argument("input_path", type=list_dir_common_check, help="Directory containing analyzed results")
    parser.add_argument("golden_path", type=list_dir_common_check, help="Directory containing analyzed results")
    parser.add_argument(
        "--output-path",
        type=check_output_path_valid,
        default=os.path.join(os.getcwd(), 'compare_result'),
        help="Output Directory after comparing."
    )
    parser.add_argument(
        '--log-level',
        type=str,
        default='info',
        choices=['debug', 'info', 'warning', 'error', 'fatal', 'critical'],
        help='Log level to print.'
    )
    parser.set_defaults(func=main)


def main(args):
    from ms_service_profiler.utils.log import set_log_level, logger
    from ms_service_profiler.utils.file_open_check import ms_open
    from msserviceprofiler.ms_service_profiler_ext.compare_tools.collector import FileCollector

    set_log_level(args.log_level)
    
    result_prefix = os.path.join(args.output_path, 'compare_result')
    file_collector = FileCollector(
        pattern=re.compile(r'(batch|service|request)_summary\.csv|profiler\.db'),
        max_iter=100
    )

    file_pairs = file_collector.collect_pairs(args.input_path, args.golden_path)
    if not file_pairs:
        logger.warning("No files to compare, please check the input directories")
        return
    
    with ms_open(f'{result_prefix}.db', 'w', encoding='utf-8'):
        with ms_open(f'{result_prefix}.xlsx', 'w', encoding='utf-8'):
            process_files(file_pairs, f'{result_prefix}.db', f'{result_prefix}.xlsx')
    
    logger.info("Comparing finished successfully, the results stored under %r", args.output_path)
    logger.info("\nWhat's Next?\n\tYou may use the `grafana` to have a better visualization of the comparison results")



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import csv
import shlex
from pathlib import Path
from typing import Optional, List, Tuple

import pandas as pd
from loguru import logger

from msserviceprofiler.modelevalstate.config.config import (
    BenchMarkConfig,
    DataStorageConfig,
    RUN_TIME,
    PerformanceIndex,
    OptimizerConfigField
)


class DataStorage:
    def __init__(self, config: DataStorageConfig):
        self.config = config
        if not self.config.store_dir.exists():
            self.config.store_dir.mkdir(parents=True)
        self.save_file = self.config.store_dir.joinpath(f"data_storage_{RUN_TIME}.csv")

    @staticmethod
    def load_history_position(load_dir: Path) -> Optional[List]:
        if not load_dir.exists():
            raise FileNotFoundError(f"file: {load_dir}")
        if not load_dir.is_dir():
            raise ValueError(f"Expect a directory, not a file.")
        history_data = []
        for file in sorted([f for f in load_dir.iterdir() if f.is_file()], key=lambda x: x.stat().st_ctime):
            if file.name.startswith("data_storage") and file.suffix == ".csv":
                data = pd.read_csv(file).to_dict(orient="records")
                history_data.extend(data)
        if not history_data:
            return None
        return history_data

    def save(self, performance_index: PerformanceIndex, params: Tuple[OptimizerConfigField],
             bench_mark_config: BenchMarkConfig, **kwargs):
        logger.info("Save result with DataStorage.")
        _column = []
        _value = []
        for k, v in performance_index.model_dump().items():
            _column.append(k)
            _value.append(v)
        for _p in params:
            _column.append(_p.name)
            _value.append(_p.value)
        benchmark_param = shlex.split(bench_mark_config.command)[2:]
        for i in range(0, len(benchmark_param), 2):
            if (i + 1) < len(benchmark_param):
                _column.append(benchmark_param[i].strip("--"))
                _value.append(benchmark_param[i + 1])
            else:
                logger.warning(f"IndexError. index: {i + 1}, list: {benchmark_param}")
        for k, v in kwargs.items():
            _column.append(k)
            _value.append(v)
        if self.save_file.exists():
            with open(self.save_file, "a+") as f:
                data_writer = csv.writer(f)
                data_writer.writerow(_value)
        else:
            with open(self.save_file, "w") as f:
                data_writer = csv.writer(f)
                data_writer.writerow(_column)
                data_writer.writerow(_value)



# -*- coding: utf-8 -*-
# Copyright (c) 2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib
import sys

from loguru import logger
from packaging import version

from msserviceprofiler.modelevalstate.inference.simulate import Simulate, ServiceField


def generate_token(self, input_metadata):
    from mindie_llm.modeling.backend_type import BackendType
    from mindie_llm.utils.env import ENV
    from mindie_llm.utils.prof.profiler import span_start, span_end, span_req

    logger.info("proprocess")
    prof = span_start("preprocess", True)
    self.plugin_data_param.q_len = None
    self.plugin_data_param.mask = None
    cache_ids, model_inputs, sampling_metadata, trace_ids = self.preprocess(input_metadata)
    logger.info("model_inputs_update_manager")
    model_inputs, qlen, mask = self.model_inputs_update_manager(model_inputs, input_metadata, cache_ids)
    self.plugin_data_param.q_len = qlen if qlen is not None else self.plugin_data_param.q_len
    self.plugin_data_param.mask = mask if mask is not None else self.plugin_data_param.mask
    span_req(prof, trace_ids)
    prof.attr("blocks", [int(x) for x in np.count_nonzero(input_metadata.block_tables > -1, axis=1)])
    logger.info("generate_token")
    logger.info("simulate init")
    try:
        Simulate.init(self)
    except Exception as e:
        logger.error(f"Failed in simulate init. error {e}")
        raise e
    try:
        Simulate.generate_features(self, input_metadata, cache_ids)
    except Exception as e:
        logger.error(f"Failed in generate features, error {e}")
        raise e
    prof = span_start("forward", True)
    logger.info("forward")
    if ServiceField.batch_field:
        Simulate.predict_and_save()
        result = Simulate.generate_logits(input_metadata.block_tables.shape[0],
                                          self.model_wrapper.config.vocab_size, self.model_wrapper.device)
    else:
        if ENV.framework_backend == BackendType.ATB:
            if (self.plugin_list and "mtp" not in self.plugin_list) or self.is_mix_model:
                result = self.generator_backend.forward(model_inputs, q_lens=self.plugin_data_param.q_len,
                                                        attn_mask=self.plugin_data_param.mask)  # q_len spec_mask
            # old graph forward
            else:
                result = self.generator_backend.forward(model_inputs, q_lens=self.plugin_data_param.q_len,
                                                        spec_mask=self.plugin_data_param.mask)
        else:
            result = self.generator_backend.forward(model_inputs, q_lens=self.plugin_data_param.q_len,
                                                    spec_mask=self.plugin_data_param.mask)  # q_len spec_mask
    if isinstance(result, tuple):
        logits, hidden_states = result
    else:
        logits = result
    span_end(prof, True)
    logger.info("sample")
    prof = span_start("sample", True)
    draft_filtered_logits = self.sample_preprocess_manager(logits, sampling_metadata, input_metadata)
    sampling_output = self.generator_backend.sample(draft_filtered_logits, sampling_metadata)
    span_end(prof, True)
    logger.info("postprocess")
    prof = span_start("postprocess", True)
    generation_output = self.postprocess(
        cache_ids, input_metadata, result, sampling_metadata, sampling_output)
    span_end(prof, True)
    generation_output.trace_ids = trace_ids
    return generation_output


def update_plugin_manager(module):
    logger.info(f"Patch {module}")
    origin_class = getattr(module, "PluginManager")
    origin_class.generate_token = generate_token
    origin_class.modelevalstate = True
    return module


def update_plugin_manager_class(module):
    logger.info(f"Patch {module}")
    module.generate_token = generate_token
    module.modelevalstate = True
    return module


_post_import_hooks = {
    "mindie_llm.text_generator.plugins.plugin_manager": update_plugin_manager,
    "mindie_llm.text_generator.plugins.plugin_manager.PluginManager": update_plugin_manager_class

}


class PostImportFinder:
    def __init__(self):
        self.skip = set()

    def find_module(self, fullname, path=None):
        if fullname in self.skip:
            return None
        if fullname not in _post_import_hooks:
            return None
        self.skip.add(fullname)
        return PostImportLoader(self)


class PostImportLoader:
    def __init__(self, finder):
        self.finder = finder

    def load_module(self, fullname):
        importlib.import_module(fullname)
        module = sys.modules[fullname]
        if fullname in _post_import_hooks:
            _post_import_hooks[fullname](module)
        self.finder.skip.remove(fullname)
        return module


class Patch2rc1:
    mindie_llm = "2.0rc1"
    mindie_llm_low = "2.0a9"

    @staticmethod
    def check_version(target_version):
        _t_v = version.parse(target_version)
        _c_v_up = version.parse(Patch2rc1.mindie_llm)
        _c_v_low = version.parse(Patch2rc1.mindie_llm_low)
        if _c_v_low < _t_v <= _c_v_up:
            return True
        else:
            return False

    @staticmethod
    def patch():
        sys.meta_path.append(PostImportFinder())
        logger.info("Successful patch")



# !/usr/bin/python3.7
# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2023-2024. All rights reserved.
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict

_PREFILL = "prefill"
_DECODE = "decode"


class StateType(Enum):
    DEFAULT = 0
    LINE = 1


@dataclass
class State:
    prefill: int = 0
    decode: int = 0
    batch_prefill: int = 0
    batch_decode: int = 0

    def __repr__(self):
        return f"TT_{self.prefill}_{self.decode}_{self.batch_prefill}_{self.batch_decode}"

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return self.__repr__() == other.__repr__()

    def __ne__(self, other):
        return self.__repr__() != other.__repr__()

    def sum(self):
        return self.prefill + self.decode + self.batch_prefill + self.batch_decode


def computer_speed(line_node, field):
    return 1 / (getattr(line_node, field) * 10 ** -3)


def computer_speed_with_second(line_node, field):
    return 1 / (getattr(line_node, field) * 10 ** -6)


def my_std(nums):
    n = len(nums)
    avg = sum(nums) / n
    return (sum(map(lambda e: (e - avg) * (e - avg), nums)) / n) ** 0.5


def get_train_sub_path(base_path: Path = Path(r"D:\PyProject\state_eval\tmp\pd_content\train")):
    # 给训练输出目录生成新的目录
    if not base_path.exists():
        base_path.mkdir(parents=True, exist_ok=True)
    _sub_len = len([0 for _ in base_path.iterdir()])
    _sub_dir = base_path.joinpath(f"{_sub_len + 1}")
    _sub_dir.mkdir(parents=True, exist_ok=True)
    return _sub_dir


def update_global_coefficient(global_coefficient: Dict, key: State, value: float) -> None:
    if key not in global_coefficient:
        global_coefficient[key] = [value]
    else:
        global_coefficient[key].append(value)


def get_module_version(module_name):
    try:
        # 方法1：直接导入模块
        # fix
        import importlib
        module = importlib.import_module(module_name)
        if hasattr(module, "__version__"):
            return module.__version__
        elif hasattr(module, "version"):
            return module.version
    except ImportError:
        pass

    try:
        # 方法2：使用 pkg_resources
        import pkg_resources
        return pkg_resources.get_distribution(module_name).version
    except (ImportError, pkg_resources.DistributionNotFound):
        pass

    try:
        # 方法3：使用 importlib.metadata（Python 3.8+）
        import importlib.metadata
        return importlib.metadata.version(module_name)
    except (ImportError):
        pass

    # 方法4：最后尝试 pip show
    try:
        import subprocess
        output = subprocess.check_output(
            ["/usr/bin/pip", "show", module_name],
            universal_newlines=True
        )
        for line in output.splitlines():
            if line.startswith("Version:"):
                return line.split(":")[1].strip()
    except subprocess.CalledProcessError:
        pass

    raise ValueError("模块未安装或无法获取版本")



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
import inspect
import logging


class HookHelper:
    def __init__(self, ori_function_define, new_function):
        self.new_function = new_function
        self.ori_function = None
        self.location = None
        self.attr_name = None
        if ori_function_define is None:
            return
        elif inspect.isfunction(ori_function_define) or inspect.ismethod(ori_function_define):
            self.ori_function = ori_function_define
            self.location, self.attr_name = self.get_location(self.ori_function)
        elif callable(ori_function_define):
            self.ori_function = ori_function_define.__call__
            self.location, self.attr_name = self.get_location(self.ori_function)

        if not all((self.ori_function, self.location, self.attr_name, self.new_function)):
            warn_msg = "{} replace failed.".format(ori_function_define)
            logging.warning(warn_msg)
            raise ValueError(warn_msg)

    @staticmethod
    def get_location(function_ins):
        if not hasattr(function_ins, "__module__"):
            warning_msg = "function {} do not has attr __module__.".format(str(function_ins))
            logging.warning(warning_msg)
            raise ValueError(warning_msg)
        module = importlib.import_module(function_ins.__module__)
        qualified_name = function_ins.__qualname__.split(".")
        classes = qualified_name[:-1]
        attr_name = qualified_name[-1]
        location = module
        for class_name in classes:
            location = getattr(location, class_name, None)
            if location is None:
                break

        if location is None:
            warning_msg = "{} do not exists".format(".".join(classes))
            logging.warning(warning_msg)
            raise ValueError(warning_msg)
        return location, attr_name

    def replace(self):
        if all((self.ori_function, self.location, self.attr_name, self.new_function)):
            setattr(self.location, self.attr_name, self._get_method(self.new_function))

    def recover(self):
        if all((self.ori_function, self.location, self.attr_name, self.new_function)):
            setattr(self.location, self.attr_name, self._get_method(self.ori_function))

    def _get_method(self, func):
        if inspect.isclass(self.location):
            func_cls_name = inspect.getattr_static(self.location, self.attr_name).__class__.__name__
            if func_cls_name in ("staticmethod", "classmethod"):
                return staticmethod(func)
        return func



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import traceback

import functools
from abc import abstractmethod
from packaging.version import Version

from .hook_helper import HookHelper


class VLLMHookerBase:
    @staticmethod
    def get_parents_name(ori_func, index=1):
        gen = traceback.walk_stack(None)
        try:
            for _ in range(index + 1):
                f = next(gen)
            return f[0].f_code.co_name
        except StopIteration:
            return None

    @abstractmethod
    def init(self):
        pass

    def do_hook(self, hook_points, profiler_func_maker, pname=None):
        for ori_func in hook_points:
            profiler_func = profiler_func_maker(ori_func)

            def replace_func(ori_func, pname):
                @functools.wraps(ori_func)
                def wrapper(*args, **kwargs):
                    if pname is not None and self.get_parents_name(ori_func) != pname:
                        return ori_func(*args, **kwargs)
                    return profiler_func(*args, **kwargs)

                return wrapper

            HookHelper(ori_func, replace_func(ori_func, pname)).replace()

    def support_version(self, version):
        if hasattr(self, "vllm_version"):
            min_version = self.vllm_version[0]
            max_version = self.vllm_version[1]
            if min_version is not None and Version(min_version) > Version(version):
                return False
            if max_version is not None and Version(max_version) < Version(version):
                return False
        return True



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from collections import namedtuple
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional
from warnings import warn

import pandas as pd

from msserviceprofiler.modelevalstate.inference.common import get_bins_and_label
from msserviceprofiler.modelevalstate.inference.file_reader import StaticFile

HARDWARE_FIELD = ("cpu_count", "cpu_mem", "soc_name", "npu_mem")
HardWare = namedtuple("HardWare", HARDWARE_FIELD, defaults=[0, 0, "", 0])

ENV_FIELD = (
    "atb_llm_razor_attention_enable", "atb_llm_razor_attention_rope", "bind_cpu", "mies_use_mb_swapper",
    "mies_pecompute_threshold",
    "mies_tokenizer_sliding_window_size", "atb_llm_lcoc_enable", "lccl_deterministic",
    "hccl_deterministic", "atb_matmul_shuffle_k_enable")
EnvField = namedtuple("EnvField", ENV_FIELD)

MINDIE_FIELD = (
    "cache_block_size", "mindie__max_seq_len", "world_size", "cpu_mem_size", "npu_mem_size", "max_prefill_tokens",
    "max_prefill_batch_size", "max_batch_size")
MindieConfig = namedtuple("MindieConfig", MINDIE_FIELD)

MODEL_CONFIG_FIELD = (
    "architectures", "hidden_act", "initializer_range", "intermediate_size", "max_position_embeddings", "model_type",
    "num_attention_heads", "num_hidden_layers", "tie_word_embeddings", "torch_dtype", "use_cache", "vocab_size",
    "quantize", "quantization_config")

ModelConfig = namedtuple("ModelConfig", MODEL_CONFIG_FIELD)


BATCH_FIELD = ("batch_stage", "batch_size", "total_need_blocks", "total_prefill_token", "max_seq_len", \
               "model_execute_time")
BATCH_FILE_FIELD = ("ibis_batchid", *BATCH_FIELD, "req_info")
BatchField = namedtuple("BatchField", BATCH_FIELD)
BatchFileField = namedtuple("BatchFileField", BATCH_FILE_FIELD)

REQUEST_FIELD = ("input_length", "need_blocks", "output_length")
REQUEST_FILE_FIELD = ("ibis_reqid", "execute_id", *REQUEST_FIELD)
RequestField = namedtuple("RequestField", REQUEST_FIELD)
RequestFileField = namedtuple("RequestFileField", REQUEST_FILE_FIELD)

MODEL_OP_FIELD = (
    "op_name", "call_count", "input_count", "input_dtype", "input_shape", "output_count", "output_dtype",
    "output_shape", "host_setup_time", "host_execute_time", "kernel_execute_time", "aic_cube_fops", "aiv_vector_fops")
ModelOpField = namedtuple("ModelOpField", MODEL_OP_FIELD)
BATCH_SIZE = "batch_size"
MAX_SEQ_LEN = "max_seq_len"

MODEL_STRUCT_FIELD = (
    "total_param_num", "total_param_size", "embed_tokens_param_size_rate", "self_attn_param_size_rate",
    "mlp_param_size_rate", "input_layernorm_param_size_rate", "post_attention_layernorm_param_size_rate",
    "norm_param_size_rate",
    "lm_head_param_size_rate")
ModelStruct = namedtuple("ModelStruct", MODEL_STRUCT_FIELD, defaults=[0 for i in range(len(MODEL_STRUCT_FIELD))])

QUESTION_FIELD = ("question", "answer")
QuestionField = namedtuple("QuestionField", QUESTION_FIELD)


class HistInfo:
    input_length = get_bins_and_label("input_length", interval=80)
    need_blocks = get_bins_and_label("need_blocks", interval=1)
    need_slots = get_bins_and_label("need_slots", interval=128)
    output_length = get_bins_and_label("output_length", interval=10, )


@dataclass
class ModelFilePaths(StaticFile):
    base_path: Path = Path("data/model")
    batch_path: Optional[Path] = None
    request_path: Optional[Path] = None

    def __post_init__(self):
        super().__post_init__()
        if self.batch_path is None:
            self.batch_path = self.base_path.joinpath("batch_need.csv")
        if self.request_path is None:
            self.request_path = self.base_path.joinpath("request_need.csv")
        for file in [self.batch_path, self.request_path]:
            if not file.exists():
                raise FileNotFoundError(file)


class FileReader:
    def __init__(self, file_paths: List[Path], num_lines: int = math.inf, start_lines: int = 0,
                 start_file_index: int = 0,
                 columns: Optional[List[str]] = None):
        self.file_paths = file_paths
        self.num_lines = num_lines
        self.current_file_index = start_file_index
        self.current_line_index = start_lines
        for _file in file_paths:
            if not _file.exists():
                raise FileNotFoundError(_file)
        self.columns = columns

    def read_rows_number(self, lines: List[pd.DataFrame]) -> int:
        if not lines:
            return self.num_lines
        rows_number = 0
        for _df in lines:
            rows_number += _df.shape[0]
        return self.num_lines - rows_number

    def read_lines(self) -> pd.DataFrame:
        lines = []
        while len(lines) < self.num_lines:
            try:
                if self.current_file_index >= len(self.file_paths):
                    # 读取完所有文件结束
                    break
                file_path = self.file_paths[self.current_file_index]
                if math.isclose(self.num_lines, math.inf):
                    df = pd.read_csv(file_path, skiprows=self.current_line_index)
                    lines.append(df)
                    # 继续读取下一个文件
                    self.current_file_index += 1
                    self.current_line_index = 0
                else:
                    _expect_nrows = self.read_rows_number(lines)
                    df = pd.read_csv(file_path, nrows=_expect_nrows, skiprows=self.current_line_index)
                    if self.columns:
                        df.columns = self.columns
                    else:
                        self.columns = df.columns.tolist()
                    lines.append(df)
                    lines_rows = sum([k.shape[0] for k in lines])
                    if lines_rows == self.num_lines:
                        # 读取到所需行结束
                        self.current_line_index += _expect_nrows
                        break
                    else:
                        # 未读取到所需行数据
                        self.current_line_index = 0
                        self.current_file_index += 1
            except Exception as e:
                warn(f"读取文件 {self.file_paths[self.current_file_index]} 时发生错误: {e}. 请核对。暂时跳过读取该文件的数据。")
                self.current_file_index += 1
                self.current_line_index = 0
        if not lines and self.current_file_index >= len(self.file_paths):
            raise StopIteration
        elif not lines:
            raise ValueError(f"lines is empty. lines: {lines}")
        return pd.concat(lines)




# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import functools

from .constraints import InvalidParameterError, make_constraint


def validate_parameter_constraint(parameter_constraint, params, caller_name):
    for param_name, param_val in params.items():
        if param_name not in parameter_constraint:
            continue

        constraint = parameter_constraint[param_name]

        if constraint == "no_validation":
            continue
        
        # support func | \ & constraint, constraint | \ & func
        constraint = make_constraint(constraint)

        if not constraint.is_satisfied_by(param_val):
            raise InvalidParameterError(
                f"The parameter '\033[0;31m{param_name!r}\033[0m' of '\033[0;31m{caller_name!r}\033[0m' "
                f"must satisfy the following requirement:\n\t \033[0;33m{constraint}\033[0m.\n"
                f"Got \033[0;31m{param_val!r}\033[0m instead."
            )


def validate_params(parameter_constraint):

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            func_sig = inspect.signature(func)

            params = func_sig.bind(*args, **kwargs)
            params.apply_defaults()

            ignore_params = [
                p.name
                for p in func_sig.parameters.values()
                if p.kind in (p.VAR_POSITIONAL, p.VAR_KEYWORD)
            ] + ["self", "cls"]
            params = {k: v for k, v in params.arguments.items() if k not in ignore_params}

            validate_parameter_constraint(
                parameter_constraint, params, caller_name=func.__qualname__
            )

            return func(*args, **kwargs)
        return wrapper

    return decorator



# -*- coding: utf-8 -*-
# Copyright (c) 2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from warnings import warn

from loguru import logger

from msserviceprofiler.modelevalstate.common import get_module_version

MINDIE_LLM = "mindie_llm"
VLLM_ASCEND = "vllm_ascend"

simulate_patch = []
optimize_patch = []
collection_patch = []
simulate_patch_elegant = []
optimize_patch_elegant = []
collection_patch_elegant = []
vllm_simulate_patch = []
vllm_optimize_patch = []

env_patch = {
    "MODEL_EVAL_STATE_COLLECT": collection_patch,
    "MODEL_EVAL_STATE_SIMULATE": simulate_patch,
    "MODEL_EVAL_STATE_ALL": optimize_patch,
    "MODEL_EVAL_STATE_COLLECT_ELEGANT": collection_patch_elegant,
    "MODEL_EVAL_STATE_SIMULATE_ELEGANT": simulate_patch_elegant,
    "MODEL_EVAL_STATE_ALL_ELEGANT": optimize_patch_elegant
}

vllm_env_patch = {
    "MODEL_EVAL_STATE_SIMULATE": vllm_simulate_patch,
    "MODEL_EVAL_STATE_ALL": vllm_optimize_patch
}

try:
    from msserviceprofiler.modelevalstate.patch.patch_manager import Patch2rc1

    simulate_patch.append(Patch2rc1)
    optimize_patch.append(Patch2rc1)
    collection_patch.append(Patch2rc1)
except ImportError as e:
    warn(f"Failed from .patch_manager import Patch2rc1. error: {e}")

try:
    from msserviceprofiler.modelevalstate.patch.plugin_simulate_patch import Patch2rc1

    simulate_patch_elegant.append(Patch2rc1)
    optimize_patch_elegant.append(Patch2rc1)
    collection_patch_elegant.append(Patch2rc1)
except ImportError as e:
    warn(f"Failed from .patch_manager import Patch2rc1. error: {e}")

try:
    from msserviceprofiler.modelevalstate.patch.patch_vllm import PatchVllm

    vllm_optimize_patch.append(PatchVllm)
    vllm_simulate_patch.append(PatchVllm)
except ImportError as e:
    warn(f"Failed from .patch_vllm import PatchVllm. error: {e}")


def enable_patch(target_env):
    flag = []
    try:
        mindie_llm_version = get_module_version(MINDIE_LLM)

        for _p in env_patch.get(target_env, []):
            if _p.check_version(mindie_llm_version):
                _p.patch()
                flag.append(_p)
    except (ModuleNotFoundError, ValueError):
        pass

    try:
        vllm_ascend_version = get_module_version(VLLM_ASCEND)
        for _p in vllm_env_patch.get(target_env):
            if _p.check_version(vllm_ascend_version):
                _p.patch()
                flag.append(_p)
    except (ModuleNotFoundError, ValueError):
        pass

    if flag:
        logger.info(f"Installed patch list {flag}.")



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
 
from loguru import logger
 
from msserviceprofiler.modelevalstate.patch.patch_manager import check_flag, add_patch
 
_patch_dir = Path(__file__).absolute().expanduser().parent.resolve()
 
 
class PatchVllm:
 
    @staticmethod
    def check_version(target_version):
        return True
 
    @staticmethod
    def patch():
        import vllm_ascend
        file_path = vllm_ascend.__path__[0]
        # 检查文件是否存在
        model_runner_file = Path(file_path).joinpath("worker/model_runner.py").resolve()
        if not model_runner_file.exists():
            raise FileNotFoundError(model_runner_file)
        plugin_manager_patch = _patch_dir.joinpath("model_runner_patch.patch")
        diff_flag = check_flag(model_runner_file, plugin_manager_patch)
        if not diff_flag:
            # 已经打过补丁，不需要打了
            logger.info("The patch already exists.")
            return
        add_patch(model_runner_file, plugin_manager_patch)




# -*- coding: utf-8 -*-
# Copyright (c) 2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import stat
from pathlib import Path

from loguru import logger
from packaging import version

_patch_dir = Path(__file__).absolute().expanduser().parent.resolve()


def check_flag(target_file, patch_file):
    with open(target_file, "r", encoding="utf-8") as f:
        data = f.readlines()
    with open(patch_file, "r", encoding="utf-8") as f:
        patch_data = f.readlines()
    i = 0
    diff_flag = True
    for _o_row in data:
        # 原来的代码
        if i == 0 and _o_row != patch_data[0]:
            continue
        if i >= len(patch_data):
            break
        # 发现有补丁代码
        if _o_row == patch_data[i]:
            i += 1
            diff_flag = False
        else:
            diff_flag = True
    return diff_flag
 
 
def add_patch(target_file, patch_file):
    flags = os.O_WRONLY | os.O_CREAT
    modes = stat.S_IWUSR | stat.S_IRUSR
    with open(patch_file, "r", encoding="utf-8") as f:
        patch_data = f.readlines()
    with os.fdopen(os.open(target_file, flags, modes), "a") as f:
        for _row in patch_data:
            f.write(_row)
    # 没有打补丁的，添加补丁文件内容
    logger.info("The patch is installed successfully.")


class Patch2rc1:
    mindie_llm = "2.0"
    mindie_llm_low = "2.0a9"

    @staticmethod
    def check_version(target_version):
        _t_v = version.parse(target_version)
        _c_v_up = version.parse(Patch2rc1.mindie_llm)
        _c_v_low = version.parse(Patch2rc1.mindie_llm_low)
        if _c_v_low < _t_v <= _c_v_up:
            pass
        else:
            logger.warning("The version may not match.")
        return True

    @staticmethod
    def patch():
        import mindie_llm
        file_path = mindie_llm.__path__[0]
        # 检查文件是否存在
        plugin_manager_file = Path(file_path).joinpath("text_generator/plugins/plugin_manager.py").resolve()
        if not plugin_manager_file.exists():
            raise FileNotFoundError(plugin_manager_file)
        plugin_manager_patch = _patch_dir.joinpath("plugin_manager_patch.patch")
        diff_flag = check_flag(plugin_manager_file, plugin_manager_patch)
        if not diff_flag:
            # 已经打过补丁，不需要打了
            logger.info("The patch aleady exists.")
            return
        add_patch(plugin_manager_file, plugin_manager_patch)




# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ms_service_profiler import Profiler, Level

from .vllm_hooker_base import VLLMHookerBase


class KVCacheManagerHook(VLLMHookerBase):
    vllm_version = ("0.6.3", "0.8.4")

    def init(self):
        from vllm.core.block_manager import SelfAttnBlockSpaceManager
        from vllm.engine.llm_engine import LLMEngine

        self.request_dict = {}

        def allocate_maker(ori_func):
            def allocate(this, seq_group, *args, **kwargs):
                profiler = Profiler(Level.INFO)
                ori_func(this, seq_group, *args, **kwargs)
                self.request_dict[seq_group.request_id] = seq_group.seqs
                prof = profiler.domain("KVCache").res(seq_group.request_id)
                prof.metric("deviceBlock", len(this.block_tables)).event("Allocate")

            return allocate

        def append_slots_maker(ori_func):
            def append_slots(this, seq, num_lookahead_slots, *args, **kwargs):
                profiler = Profiler(Level.INFO)
                request_id = seq.seq_id
                for k, v in self.request_dict.items():
                    if seq in v:
                        request_id = k
                        break
                new_cows = ori_func(this, seq, num_lookahead_slots, *args, **kwargs)
                num_blocks = len(getattr(this.block_tables.get(seq.seq_id), "_blocks", []))
                prof = profiler.domain("KVCache").res(request_id)
                prof.metric("deviceBlock", len(this.block_tables)).event("AppendSlot")
                # This is called for every decoder process, recording current blocks
                profiler.domain("KVCache").res(request_id).metric("deviceBlock", num_blocks).event("blocks")
                return new_cows

            return append_slots

        def swap_in_maker(ori_func):
            def swap_in(this, seq_group, *args, **kwargs):
                profiler = Profiler(Level.INFO)
                res = ori_func(this, seq_group, *args, **kwargs)
                prof = profiler.domain("KVCache").res(seq_group.request_id)
                prof.attr("swap", "swap_in").metric("deviceBlock", len(this.block_tables)).event("SwapIn")
                return res

            return swap_in

        def swap_out_maker(ori_func):
            def swap_out(this, seq_group, *args, **kwargs):
                profiler = Profiler(Level.INFO)
                res = ori_func(this, seq_group, *args, **kwargs)
                prof = profiler.domain("KVCache").res(seq_group.request_id)
                prof.attr("swap", "swap_out").metric("deviceBlock", len(this.block_tables)).event("SwapOut")
                return res

            return swap_out

        def free_maker(ori_func):
            def free(this, seq, *args, **kwargs):
                profiler = Profiler(Level.INFO)
                request_id = seq.seq_id
                for k, v in self.request_dict.items():
                    if seq in v:
                        request_id = k
                        break
                ori_func(this, seq, *args, **kwargs)
                profiler.domain("KVCache").res(request_id).metric("deviceBlock", len(this.block_tables)).event("Free")

            return free

        def get_stats_maker(ori_func):
            def get_stats(this, *args, **kwargs):
                profiler = Profiler(Level.INFO)
                stats = ori_func(this, *args, **kwargs)
                num_free_gpu = sum(scheduler.block_manager.get_num_free_gpu_blocks() for scheduler in this.scheduler)
                profiler.domain("KVCache").attr("cpuHitCache", stats.cpu_cache_usage_sys)
                profiler.attr("hitCache", stats.gpu_cache_usage_sys).event("GetCacheHitRate")
                profiler.domain("KVCache").attr("deviceFreeBlock", num_free_gpu).event("GetCacheHitRate")
                return stats

            return get_stats

        self.do_hook([SelfAttnBlockSpaceManager.allocate], allocate_maker)
        self.do_hook([SelfAttnBlockSpaceManager.append_slots], append_slots_maker)
        self.do_hook([SelfAttnBlockSpaceManager.swap_in], swap_in_maker)
        self.do_hook([SelfAttnBlockSpaceManager.swap_out], swap_out_maker)
        self.do_hook([SelfAttnBlockSpaceManager.free], free_maker)
        self.do_hook([LLMEngine._get_stats], get_stats_maker)


kvcache_hookers = [KVCacheManagerHook]



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


CONFIG_FILE_MAX_SIZE = 1 * 1024 * 1024 # work for .ini config file
TEXT_FILE_MAX_SIZE = 5 * 1024 * 1024 * 1024 # work for txt, csv, py
JSON_FILE_MAX_SIZE = 1024 * 1024 * 1024
DB_MAX_SIZE = 50 * 1024 * 1024 * 1024
LOG_FILE_MAX_SIZE = 5 * 1024 * 1024


EXT_SIZE_MAPPING = {
    '.db': DB_MAX_SIZE,
    ".ini": CONFIG_FILE_MAX_SIZE,
    '.csv': TEXT_FILE_MAX_SIZE,
    '.txt': TEXT_FILE_MAX_SIZE,
    '.json': JSON_FILE_MAX_SIZE,
    '.log': LOG_FILE_MAX_SIZE
}



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import time

from ms_service_profiler import Profiler, Level
from .vllm_hooker_base import VLLMHookerBase

GLOBAL_FORWARD_PROF = []


def prof_add_request(request_id, prompt, *args, **kwargs):
    # 记录请求进入系统的时间
    Profiler(Level.INFO).domain("Request").res(request_id).event("httpReq")
    Profiler(Level.INFO).domain("Request").res(request_id).event("encode")


# generate -> add_request -> schedule -> execute_model
# 在请求进入引擎时记录时间戳，用于后续计算队列等待时间。
class EngineRequestTrackerHook063(VLLMHookerBase):
    vllm_version = ("0.6.3", "0.6.3")

    def init(self):
        from vllm.engine.llm_engine import LLMEngine
        from vllm.engine.async_llm_engine import AsyncLLMEngine

        def add_request_maker(ori_func):
            def add_request(this, request_id, prompt, *args, **kwargs):
                prof_add_request(request_id, prompt, *args, **kwargs)
                return ori_func(this, request_id, prompt, *args, **kwargs)

            return add_request

        self.do_hook([LLMEngine.add_request, AsyncLLMEngine.add_request], add_request_maker)


class EngineRequestTrackerHook084(VLLMHookerBase):
    vllm_version = ("0.8.4", "0.8.4")

    def init(self):
        from vllm.engine.llm_engine import LLMEngine
        from vllm.engine.async_llm_engine import AsyncLLMEngine

        def add_request_maker(ori_func):
            def add_request(this, request_id, prompt, *args, **kwargs):
                prof_add_request(request_id, prompt, *args, **kwargs)
                return ori_func(this, request_id, prompt, *args, **kwargs)

            return add_request

        def add_request_async_maker(ori_func):
            async def add_request(this, request_id, prompt, *args, **kwargs):
                prof_add_request(request_id, prompt, *args, **kwargs)
                return ori_func(this, request_id, prompt, *args, **kwargs)

            return add_request

        self.do_hook([LLMEngine.add_request], add_request_maker)
        self.do_hook([AsyncLLMEngine.add_request], add_request_async_maker)


# 采集请求返回的数据
class LLMEngineHook084(VLLMHookerBase):
    vllm_version = ("0.8.4", "0.8.4")

    def init(self):
        from vllm.engine.llm_engine import LLMEngine

        def process_model_outputs_maker(ori_func):
            def process_model_outputs(this, ctx, request_id=None, *args, **kwargs):
                if len(ctx.output_queue) == 0:
                    return ori_func(this, ctx, request_id, *args, **kwargs)

                request_id_list = []
                outputs, metadata_list, scheduler_outputs, _, _, _, skip = ctx.output_queue[0]
                for ii, (sch_seq_group, meta) in enumerate(zip(scheduler_outputs.scheduled_seq_groups, metadata_list)):
                    if ii in skip:
                        continue
                    seq_group, seq_request_id = sch_seq_group.seq_group, meta.request_id
                    request_id_list.append(seq_request_id)
                    if seq_group.is_finished() and len(seq_group.seqs) > 0:
                        cur_seq = seq_group.seqs[0]
                        profiler_recv = Profiler(Level.INFO).domain("Request").res(seq_request_id)
                        profiler_reply = Profiler(Level.INFO).domain("Request").res(seq_request_id)
                        profiler_recv.metric("recvTokenSize", cur_seq.get_prompt_len()).event("httpRes")
                        profiler_reply.metric("replyTokenSize", cur_seq.get_output_len()).event("httpRes")

                prof = Profiler(Level.INFO).domain("Request").res(request_id_list)
                ret = ori_func(this, ctx, request_id, *args, **kwargs)
                prof.event("DecodeEnd")
                return ret

            return process_model_outputs

        self.do_hook([getattr(LLMEngine, "_process_model_outputs")], process_model_outputs_maker)


class LLMEngineHook063(VLLMHookerBase):
    vllm_version = ("0.6.3", "0.6.3")

    def init(self):
        from vllm.engine.llm_engine import LLMEngine

        def validate_output_maker(ori_func):
            def validate_output(output, output_type):
                profiler_recv = Profiler(Level.INFO).domain("Request")
                profiler_reply = Profiler(Level.INFO).domain("Request")
                if output.finished is True:
                    request_id = output.request_id
                    input_token_size = len(output.prompt_token_ids)
                    output_token_size = len(output.outputs[0].token_ids)
                    profiler_recv.res(request_id).metric("recvTokenSize", input_token_size).event("httpRes")
                    profiler_reply.res(request_id).metric("replyTokenSize", output_token_size).event("httpRes")
                return ori_func(output, output_type)

            return validate_output

        self.do_hook([LLMEngine.validate_output], validate_output_maker)


request_hookers = [EngineRequestTrackerHook063, EngineRequestTrackerHook084, LLMEngineHook063, LLMEngineHook084]



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass

from .logic import FunctionConstraint, IfElseConstraint
from ._path import (
    IsFile, Exists, IsDir, HasSoftLink, IsReadable, 
    IsWritable, IsExecutable, IsWritableToGroupOrOthers, 
    IsConsistentToCurrentUser, IsSizeReasonable
)
from .base import BaseConstraint


class PathConstraintBuilder:
    """
    Constraint builder for IDE support.

    Available constraints:
        is_file, file_exists, is_dir, has_soft_link, is_readable, is_writable,
        is_executable, is_not_writable_to_group_or_others, is_consistent_to_current_user,
        is_size_reasonable
    """
    @staticmethod
    def is_file(): 
        return IsFile()
    
    @staticmethod
    def file_exists(): 
        return Exists()
    
    @staticmethod
    def is_dir(): 
        return IsDir()
    
    @staticmethod
    def has_soft_link(): 
        return HasSoftLink()
    
    @staticmethod
    def is_readable(): 
        return IsReadable()
    
    @staticmethod
    def is_writable(): 
        return IsWritable()
    
    @staticmethod
    def is_executable(): 
        return IsExecutable()
    
    @staticmethod
    def is_writable_to_group_or_others():
        return IsWritableToGroupOrOthers()
    
    @staticmethod
    def is_consistent_to_current_user():
        return IsConsistentToCurrentUser()
    
    @staticmethod
    def is_size_reasonable(*, size_limit=None, require_confirm=True):
        return IsSizeReasonable(size_limit=size_limit, require_confirm=require_confirm)


Path = PathConstraintBuilder


def make_constraint(constraint, description=None):
    if isinstance(constraint, BaseConstraint):
        return constraint

    if callable(constraint):
        return FunctionConstraint(constraint, description)

    raise TypeError(
        f"Expected a BaseConstraint instance or a callable, but got {type(constraint).__name__}."
    )


def where(condition, if_constraint, else_constraint, *, description=None):
    return IfElseConstraint(
        condition,
        make_constraint(if_constraint),
        make_constraint(else_constraint),
        description=description
    )


@dataclass(frozen=True)
class Rule:
    read_file_common_check: BaseConstraint = where(
        os.getuid() == 0, 
        Path.is_file(),
        Path.is_file() & ~Path.has_soft_link() & 
        Path.is_readable() & ~Path.is_writable_to_group_or_others() & 
        Path.is_consistent_to_current_user() & Path.is_size_reasonable(),
        description="current user is root"
    )

    exec_file_common_check: BaseConstraint = where(
        os.getuid() == 0, 
        Path.is_file(),
        Path.is_file() & ~Path.has_soft_link() & 
        Path.is_writable() & ~Path.is_writable_to_group_or_others() & 
        Path.is_consistent_to_current_user() & Path.is_size_reasonable(),
        description="current user is root"
    )



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from msserviceprofiler.vllm_profiler.vllm_profiler_core.model_hookers import model_hookers
from msserviceprofiler.vllm_profiler.vllm_profiler_core.batch_hookers import batch_hookers
from msserviceprofiler.vllm_profiler.vllm_profiler_core.kvcache_hookers import kvcache_hookers
from msserviceprofiler.vllm_profiler.vllm_profiler_core.request_hookers import request_hookers

all_hookers = kvcache_hookers
all_hookers += model_hookers
all_hookers += batch_hookers
all_hookers += request_hookers

for hook_cls in all_hookers:
    hooker = hook_cls()
    if hooker.support_version("0.6.3"):
        hooker.init()



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from pathlib import Path
import logging
from collections import namedtuple

TARGETS = namedtuple("TARGETS", ["FirstTokenTime", "Throughput"])("FirstTokenTime", "Throughput")
_SUGGESTION_TYPES = ["env", "system", "config"]
SUGGESTION_TYPES = namedtuple("SUGGESTION_TYPES", _SUGGESTION_TYPES)(*_SUGGESTION_TYPES)
MAX_FILE_ITER_TIME = 10000
MAX_FILE_SIZE = 10
BYTES_TO_GB = 1024**3

LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "fatal": logging.FATAL,
    "critical": logging.CRITICAL,
}


def str_ignore_case(value):
    return value.lower().replace("_", "").replace("-", "")


def walk_dict(data, parent_key=""):
    if isinstance(data, dict):
        for key, value in data.items():
            if not isinstance(value, (dict, tuple, list)):
                yield key, value, parent_key
            else:
                new_key = f"{parent_key}.{key}" if parent_key else key
                yield from walk_dict(value, new_key)
    elif isinstance(data, (tuple, list)):
        for index, item in enumerate(data):
            if not isinstance(item, (dict, tuple, list)):
                yield str(index), item, parent_key
            else:
                new_key = f"{parent_key}.{index}" if parent_key else index
                yield from walk_dict(item, new_key)


def set_log_level(level="info"):
    if level.lower() in LOG_LEVELS:
        logger.setLevel(LOG_LEVELS.get(level.lower()))
    else:
        logger.warning("Set %s log level failed.", level)


def set_logger(msit_logger):
    msit_logger.propagate = False
    msit_logger.setLevel(logging.INFO)
    if not msit_logger.handlers:
        stream_handler = logging.StreamHandler()
        formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
        stream_handler.setFormatter(formatter)
        msit_logger.addHandler(stream_handler)


logger = logging.getLogger("msservice_advisor_logger")
set_logger(logger)


def vaild_readable_directory(path):
    if not os.path.exists(path):
        raise FileExistsError(f"Path '{path}' does not exist.")
    if not os.path.isdir(path):
        raise ValueError(f"Path '{path}' is not a directory.")
    if not os.access(path, os.R_OK):
        raise PermissionError(f"Directory '{path}' is not readable.")


def vaild_readable_file(path):
    path = Path(path)
    if not path.exists():
        raise FileExistsError(f"path '{path}' does not exist.")

    if not os.access(path, os.R_OK):
        raise PermissionError(f"path '{path}' is not readable.")
    
    file_size = path.stat().st_size
    if file_size > MAX_FILE_SIZE * BYTES_TO_GB:
        raise ValueError(f"path '{path}' cannot exceed {MAX_FILE_SIZE}GB.")
    return path


def get_directory_size(path):
    iter_time = 0
    total_size = 0
    for dirpath, _, filename in os.walk(path):
        for f in filename:
            if iter_time > MAX_FILE_ITER_TIME:
                raise ValueError(f"path '{path}' iter times cannot exceed {MAX_FILE_ITER_TIME}.")
            iter_time += 1

            fp = os.path.join(dirpath, f)
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size / BYTES_TO_GB


# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path

import pandas as pd
from ms_service_profiler.utils.log import logger

from .base import BaseComparator
from ..common.csv_fields import BaseCSVFields, ServiceCSVFields


class CSVComparator(BaseComparator):
    SUPPORTED_EXTENSIONS = ['.csv']

    def process(self, file_a, file_b):
        df = self.compare_csv(file_a, file_b)
        self._save_results(df, file_a)

    def compare_csv(self, file_a, file_b):
        try:
            df_a = pd.read_csv(file_a)
            df_b = pd.read_csv(file_b)
        except Exception as ex:
            logger.error(f'failed to read csv, please check {file_a}.')
            return pd.DataFrame()

        # 确保列名一致
        if set(df_a.columns) != set(df_b.columns):
            logger.error("两个 CSV 文件的列名不一致！")
            return pd.DataFrame()

        # 按 Metric 列合并两个 DataFrame
        df_merged = pd.merge(df_a, df_b, on=BaseCSVFields.METRIC, suffixes=('_a', '_b'))

        # 计算差值
        for col in df_a.columns[1:]:  # 跳过 Metric 列
            got_error = False
            try:
                # 计算绝对差值
                abs_diff = (df_merged[f'{col}_a'] - df_merged[f'{col}_b'])

                # 计算百分比相对差值（避免除以零错误）
                rel_diff = abs_diff / df_merged[f'{col}_a'].replace(0, pd.NA) * 100
                with pd.option_context('future.no_silent_downcasting', True):
                    rel_diff = rel_diff.fillna(0)  # 将 NaN 替换为 0

                # 合并为字符串格式
                df_merged[f'{col}_diff'] = abs_diff.round(2).astype(str) + '|' + rel_diff.round(2).astype(str) + '%'
            except Exception as ex:
                error_msg = f'Calculate Diff Error: f{ex}'
                df_merged[f'{col}_diff'] = error_msg
                if not got_error:
                    logger.warning(error_msg)
                    got_error = True

        # 存储所有行的列表
        rows = []

        # 遍历合并后的 DataFrame
        for _, row in df_merged.iterrows():
            metric = row[BaseCSVFields.METRIC]

            # 添加 a 数据
            a_row = {BaseCSVFields.METRIC: metric, 'Data Source': 'Input Data'}
            for col in df_a.columns[1:]:
                a_row[col] = row[f'{col}_a']
            rows.append(a_row)

            # 添加 b 数据
            b_row = {BaseCSVFields.METRIC: metric, 'Data Source': 'Golden Data'}
            for col in df_a.columns[1:]:
                b_row[col] = row[f'{col}_b']
            rows.append(b_row)

            # 添加 diff 数据
            diff_row = {BaseCSVFields.METRIC: metric, 'Data Source': 'Different'}
            for col in df_a.columns[1:]:
                diff_row[col] = row[f'{col}_diff']
            rows.append(diff_row)

        # 将所有行转换为 DataFrame
        result = pd.DataFrame(rows)
        return result

    def _save_visualization_database(self, df, sheet_name):
        if df.shape[0] == 0:
            return
        if sheet_name == "service":
            for i in range(0, len(df), 3):
                if i + 3 <= len(df):
                    new_col_name = df.loc[i, ServiceCSVFields.METRIC]
                    df[new_col_name] = df.iloc[i:i + 3, :][ServiceCSVFields.VALUE].reset_index(drop=True)
            df.drop(columns=[ServiceCSVFields.METRIC, ServiceCSVFields.VALUE], inplace=True)
            df = df.iloc[:3, :]

        df.to_sql(name=sheet_name, con=self.out_db_conn, if_exists='replace', index=False)

    def _save_results(self, df, source_file):
        sheet_name = Path(source_file).stem.split('_')[0]
        df.to_excel(self.excel_writer, sheet_name=sheet_name, index=False)
        self._save_visualization_database(df, sheet_name)



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ms_service_profiler import Profiler, Level
from .vllm_hooker_base import VLLMHookerBase

GLOBAL_FORWARD_PROF = []


class ExecutorBaseExecuteModelHook(VLLMHookerBase):
    vllm_version = ("0.8.4", "0.8.4")

    def init(self):
        from vllm.executor.executor_base import ExecutorBase
        from vllm.executor.executor_base import DistributedExecutorBase

        def execute_model_maker(ori_func):
            def execute_model(this, execute_model_req, *args, **kwargs):
                prof = Profiler(Level.INFO).domain("ModelExecute")
                is_prefill, request_id_list, request_id_with_iter_list = False, [], []
                for seq_metadata in execute_model_req.seq_group_metadata_list:
                    if len(seq_metadata.seq_data) > 0:
                        cur_seq_data = list(seq_metadata.seq_data.values())[0]
                        iter_size = cur_seq_data.get_len() - len(cur_seq_data.prompt_token_ids)
                    else:
                        iter_size = 0
                    request_id_list.append({"rid": seq_metadata.request_id})
                    request_id_with_iter_list.append({"rid": seq_metadata.request_id, "iter_size": iter_size})
                    is_prefill = is_prefill or seq_metadata.is_prompt

                prof.res(request_id_with_iter_list)
                prof.attr("batch_type", "Prefill" if is_prefill else "Decode")
                prof.span_start("modelExec")
                prof.attr("batch_size", len(execute_model_req.seq_group_metadata_list))

                preprocess_prof = Profiler(Level.INFO).domain("ModelExecute").res(request_id_list)
                preprocess_prof.event("preprocess")

                ret = ori_func(this, execute_model_req, *args, **kwargs)
                prof.span_end()
                return ret

            return execute_model

        self.do_hook([ExecutorBase.execute_model], execute_model_maker)
        self.do_hook([DistributedExecutorBase.execute_model], execute_model_maker)


class ModelRunnerExecuteHook(VLLMHookerBase):
    vllm_version = ("0.6.3", "0.8.4")

    def __init__(self):
        super().__init__()
        self.is_model_first_run = True

    def init(self):
        from vllm.worker.model_runner import ModelRunner

        def execute_model_maker(ori_func):
            def execute_model(this, model_input, kv_caches, *args, **kwargs):
                if self.is_model_first_run:
                    self.is_model_first_run = False
                    return ori_func(this, model_input, kv_caches, *args, **kwargs)

                prof = Profiler(Level.INFO).domain("ModelExecute")
                prof.span_start("modelExec")

                ret = ori_func(this, model_input, kv_caches, *args, **kwargs)

                is_prefill = model_input.attn_metadata.prefill_metadata

                request_id_list = []

                for request_id, _ in model_input.request_ids_to_seq_ids.items():
                    request_id_list.append({"rid": request_id})

                prof.res(request_id_list)

                if is_prefill:
                    prof.attr("batch_type", "Prefill")
                else:
                    prof.attr("batch_type", "Decode")

                batch_size = model_input.input_tokens.shape[0]
                prof.attr("batch_size", batch_size)

                prof.span_end()
                return ret

            return execute_model

        self.do_hook([ModelRunner.execute_model], execute_model_maker)


class ModelForwardHook(VLLMHookerBase):
    vllm_version = ("0.6.3", "0.8.4")

    def __init__(self):
        super().__init__()
        self.is_forward_first_run = True

    def init(self):
        from vllm.attention.backends.utils import CommonAttentionState

        def begin_forward_maker(ori_func):
            def begin_forward(this, model_input, *args, **kwargs):
                ret = ori_func(this, model_input, *args, **kwargs)
                if self.is_forward_first_run:
                    self.is_forward_first_run = False
                    return ret

                request_id_list = [{"rid": request_id} for request_id, _ in model_input.request_ids_to_seq_ids.items()]
                prof = Profiler(Level.INFO).domain("ModelExecute").res(request_id_list)
                GLOBAL_FORWARD_PROF.append(prof)
                return ret

            return begin_forward

        self.do_hook([CommonAttentionState.begin_forward], begin_forward_maker)


class SetForwardContextHook(VLLMHookerBase):
    vllm_version = ("0.8.4", "0.8.4")

    def init(self):
        from vllm import forward_context
        from contextlib import contextmanager

        def set_forward_context_maker(ori_func):
            @contextmanager
            def set_forward_context(*args, **kwargs):
                if len(GLOBAL_FORWARD_PROF) > 0:
                    prof = GLOBAL_FORWARD_PROF.pop(0)
                    prof.span_start("forward")
                else:
                    prof = None
                with ori_func(*args, **kwargs):
                    yield
                if prof is not None:
                    prof.span_end()

            return set_forward_context

        self.do_hook([forward_context.set_forward_context], set_forward_context_maker)


model_hookers = [ExecutorBaseExecuteModelHook, ModelRunnerExecuteHook, ModelForwardHook, SetForwardContextHook]



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import math
import os
import re
import subprocess

from msserviceprofiler.msservice_advisor.profiling_analyze.register import register_analyze, answer
from msserviceprofiler.msservice_advisor.profiling_analyze.utils import logger, SUGGESTION_TYPES, BYTES_TO_GB
from msserviceprofiler.msservice_advisor.profiling_analyze.utils import vaild_readable_directory, vaild_readable_file
from msserviceprofiler.msservice_advisor.profiling_analyze.utils import get_directory_size


def get_benchmark_token_num(benchmark, info_name):
    token_num = benchmark.get("result_perf").get(info_name).get('average')
    return int(float(token_num))


def extract_token_num(benchmark, input_params):
    """
        获取请求输入参数：
        input_token_num: 输入长度。优先从用户输入中获取; 用户未输入从benchmark中获取。
        avg_token_num: 平均输出长度。优先从benchmark中获取; 未获取到则与用户输入相等。
    """
    # get from benchmark
    avg_token_num = 0
    input_token_num = 0
    try:
        input_token_num = get_benchmark_token_num(benchmark, 'InputTokens')
        avg_token_num = get_benchmark_token_num(benchmark, 'GeneratedTokens')
    except Exception as e:
        logger.warning(f"Extract request input info from benchmark failed due to {e}, "
            "please check or try to specifying the input length.")

    # get from user input
    if input_params.input_token_num > 0:
        input_token_num = input_params.input_token_num
    if input_params.output_token_num > 0:
        avg_token_num = input_params.output_token_num

    logger.info(f"input_token_num: {input_token_num}, avg_output_token_num: {avg_token_num}")
    return input_token_num, avg_token_num


def get_schedule_config_info(backend_config, output_token_num):
    schedule_config = backend_config.get('ScheduleConfig')
    cache_block_sizes = schedule_config.get('cacheBlockSize')

    if not cache_block_sizes:
        raise Exception("mindie-server config.json missing 'cacheBlockSize'.")

    if output_token_num == 0:
        output_token_num = int(schedule_config.get('maxIterTimes', 0))

    logger.info(f"output_token_num: {output_token_num}, cache_block_sizes: {cache_block_sizes}")
    return output_token_num, cache_block_sizes


def get_model_config_info(model_configs, tp, npu_device_ids):
    if tp == 0:
        tp_config = model_configs.get('tp')
        tp = int(tp_config) if tp_config else len(npu_device_ids)
        tp = max(1, tp)

    model_weight_path = model_configs.get('modelWeightPath')
    if not model_weight_path:
        raise Exception("mindie-server config.json missing 'modelWeightPath'.")
    
    logger.info(f"tp: {tp}, model_weight_path: {model_weight_path}")
    return tp, model_weight_path


def extract_server_config_params(server_config, args):
    """
        获取mindie-server config.json中的参数信息
        output_token_num: 请求输出长度。优先从用户输入中获取; 用户未输入等于conf中的maxIterTimes。
        tp: tp域信息。优先从优先从用户输入中获取; 用户未输入等于conf中的tp; 如果conf未配置tp, 则默认为使用的npu卡数。
        cache_block_sizes: cache块大小,一般为128。从conf中的cacheBlockSize获取。
        npu_device_ids: 使用npu的卡号。从config中的npuDeviceIds获取。
        model_weight_path: 使用的模型权重路径, 从conf中的modelWeightPath获取。
        npu_mem_size: 单卡预留给kvcache的显存, 单位GB, 从conf中的npuMemSize获取。获取为-1则需要重新计算显存大小。
        sp: Sequence Parallelism策略, 对Sequence进行切分, 从conf中的sp获取, 若未配置默认为1。
    """
    output_token_num = args.output_token_num
    tp = args.tp

    backend_config = server_config.get('BackendConfig')
    output_token_num, cache_block_sizes = get_schedule_config_info(backend_config, output_token_num)

    npu_device_ids = backend_config.get('npuDeviceIds')[0]
    if not npu_device_ids or len(npu_device_ids) == 0:
        raise Exception("mindie-server config.json missing 'npuDeviceIds'.")
    logger.info(f"npu_device_ids: {npu_device_ids}")

    model_configs = backend_config.get('ModelDeployConfig').get('ModelConfig')[0]
    tp, model_weight_path = get_model_config_info(model_configs, tp, npu_device_ids)

    return dict(
        output_token_num=output_token_num,
        cache_block_sizes=cache_block_sizes,
        npu_device_ids=npu_device_ids,
        tp=tp,
        model_weight_path=model_weight_path,
        npu_mem_size=model_configs.get('npuMemSize', 0),
        sp=model_configs.get('sp', 1)
    )


def extract_model_config_params(model_weight_path):
    """
        获取模型参数信息，及模型权重文件总大小
    """
    # check model weight directory
    vaild_readable_directory(model_weight_path)

    # check model config.json
    model_config_file = os.path.join(model_weight_path, "config.json")
    model_config_file = vaild_readable_file(model_config_file)
    logger.info(f"model config path: {model_config_file}")

    # get model config.json
    with open(model_config_file, 'r') as f:
        model_params = json.load(f)

    # caculate model weight size
    model_weight_size = get_directory_size(model_weight_path)
    logger.info(f"model wight size: {model_weight_size}GB")

    return model_params, model_weight_size



def check_vaild_smi_output(output):
    header_line = None
    for line in output.splitlines():
        stripped_line = line.strip()
        if stripped_line.startswith("NPU ID"):
            header_line = stripped_line
            break
    if not header_line:
        raise ValueError("Header line not found.")
    
    # parse column indices
    headers = re.split(r'\s{2,}', header_line)
    col_map = {name: idx for idx, name in enumerate(headers)}

    # vaildate required columns
    requlire_cols = ["NPU ID", "Chip Logic ID"]
    for col in requlire_cols:
        if col not in col_map:
            raise ValueError(f"Missing required column: {col}")
        
    return header_line, col_map


def get_npu_ids_map():
    """
        使用 npu-smi -m 获取信息
        返回 dict {npu_id: chip_logic_id}, 其中chip_logic_id即为device id
    """
    cmd = ['npu-smi', 'info', '-m']
    try:
        output = subprocess.check_output(
            cmd, text=True, stderr=subprocess.STDOUT
        )
    except Exception as e:
        raise RuntimeError(f"Command 'npu-smi info -m' failed: {e}") from e

    header_line, col_map = check_vaild_smi_output(output)

    npu_idx = col_map["NPU ID"]
    logic_idx = col_map["Chip Logic ID"]

    npu_ids_map = {}
    for line in output.splitlines():
        # skip empty line and header
        if not line.strip() or line == header_line:
            continue

        # skip invaild colums
        cols = re.split(r'\s{2,}', line.strip())
        if len(cols) < max(npu_idx, logic_idx) + 1:
            continue

        npu_id = cols[npu_idx]
        logic_id = cols[logic_idx]

        if npu_id.isdigit() and logic_id.isdigit():
            npu_ids_map[int(logic_id)] = int(npu_id)

    logger.debug(f"npu_ids_map: {npu_ids_map}")
    return npu_ids_map


def cal_available_gm_memory(gm_capacity, gm_usage):
    if gm_capacity is None or gm_usage is None:
        return 0
    
    available_mem = gm_capacity * (1 - gm_usage / 100)
    return round(available_mem / 1024, 2)


def cal_npu_gm_memory(npu_id):
    """
        使用 npu-smi info -i {npu_id} -t usages 命令行获取单卡npu内存大小
        M Usage Rate(%): 当前npu占用内存
        M Capacity(MB): npu总内存
        返回 M Capacity * (1 - M Usage Rate)
    """
    try:
        cmd = ['npu-smi', 'info', '-i', str(npu_id), '-t', 'usages']
        output = subprocess.check_output(
            cmd, text=True, stderr=subprocess.STDOUT
        )

        gm_capacity = None
        gm_usage = None
        for line in output.splitlines():
            # Match M capacity line (e.g. "xxM Capacity(MB) : 65536")
            capacity_match = re.match(
                r".*M Capacity\(MB\)\s*:\s*(\d+)\s*$",
                line.strip()
            )
            if capacity_match:
                gm_capacity = int(capacity_match.group(1))
                continue

            # Match M Usage line (e.g. "xxM Usage Rate(%) : 5")
            usage_match = re.match(
                r".*M Usage Rate\(%\)\s*:\s*(\d+)\s*$",
                line.strip()
            )
            if usage_match:
                gm_usage = int(usage_match.group(1))
                continue

        return cal_available_gm_memory(gm_capacity, gm_usage)
    except Exception as e:
        logger.warning(f"Command npu-smi info -i $npu_id -t usages parse failed due to {e}")
        return 0


def get_available_npu_memory(npu_device_ids):
    # get device-id to npu-id map
    npu_ids_map = get_npu_ids_map()
    if not npu_ids_map:
        raise Exception(f"parse npu-smi info -m failed.")
    
    available_npu_memory = float('inf')
    for npu_device_id in npu_device_ids:
        npu_id = npu_ids_map.get(int(npu_device_id), 0)
        pre_npu_memory = cal_npu_gm_memory(npu_id)
        available_npu_memory = min(available_npu_memory, pre_npu_memory)
    
    logger.info(f"available_npu_memory: {available_npu_memory}")
    return available_npu_memory


def cal_npu_mem_size(server_params, model_weight_size):
    """
        优先从mindie-server config.json的npuMemSize获取, 若该值非法, 则使用公式计算
        计算显存公式：
        Floor[(单卡可用显存 - 权重/NPU卡数) * 系数]
    """
    # get from mindie server config, GB
    npu_mem_size = server_params['npu_mem_size']
    if npu_mem_size >= 0:
        logger.info(f"npu_mem_size: {npu_mem_size}GB")
        return npu_mem_size

    # calulate npu available memory size
    npu_available_mem = get_available_npu_memory(server_params['npu_device_ids'])

    npu_num = len(server_params['npu_device_ids'])
    if npu_num == 0:
        raise ValueError(f"used npu number cannot be 0")

    fixed_coefficent = 0.8
    npu_mem_size = (npu_available_mem - model_weight_size / npu_num) * fixed_coefficent
    npu_mem_size = max(math.floor(npu_mem_size), 0)

    logger.info(f"npu_memory_size: {npu_mem_size}GB")
    return npu_mem_size


def get_model_param(model_params, param_name):
    param_value = model_params.get(param_name, 0)
    logger.info(f"{param_name}: {param_value}")
    return param_value


def cal_total_block_num(npu_mem_size, server_params, model_params):
    """
        Total Block Num =
        Floor[npu显存/(模型网络层数*cacheBlockSize*模型注意力头数*注意力头大小*Cache类型字节数*Cache数)]
        其中, Cache数默认为2

        模型config中:
        num_hidden_layers: 模型网络层数
        num_attention_heads: 模型查询头数
        hidden_size: 模型隐藏层维度
        torch_dtype: Cache类型
        num_key_value_heads: kv键值头数
    """
    tp = server_params['tp'] # tp checked before cannot be 0
    cache_block_sizes = server_params['cache_block_sizes']

    num_hidden_layers = get_model_param(model_params, 'num_hidden_layers')
    num_attention_heads = get_model_param(model_params, 'num_attention_heads')
    hidden_size = get_model_param(model_params, 'hidden_size')
    torch_dtype = get_model_param(model_params, 'torch_dtype')
    torch_dtype_size = 1 if torch_dtype == "uint8" else 2 # bf16 type dtype byte is 2

    # for MHA models, attention_size is hidden_size
    attention_size = hidden_size / tp
    num_key_value_heads = model_params.get('num_key_value_heads')
    if num_key_value_heads:
        logger.info(f"num_key_value_heads:{num_key_value_heads}")

        # for GQA or MLA model, attention_size is num_key_value_heads plus attention_head_size
        if num_attention_heads <= 0:
            raise ValueError(f'Invaild num_attention_heads.')
        attention_head_size = hidden_size / num_attention_heads / tp
        attention_size = attention_head_size * num_key_value_heads
    logger.debug(f"attention_size:{attention_size}")

    denominator = num_hidden_layers * cache_block_sizes * attention_size * torch_dtype_size * 2
    if denominator <= 0:
        raise ValueError(f'Calculate total block num with invaild denominator.')
    
    return math.floor(npu_mem_size * BYTES_TO_GB / denominator)


def cal_block_nums(
    input_tokens: int,
    cache_block_size: int,
    output_tokens: int,
    avg_tokens: int
):
    """
        所需最大Block Num = Ceil(输入token数/cache_block_size)+Celi(最大输出token数/cache_block_size)
        所需最小Block Num = Ceil(输入token数/cache_block_size)
        所需平均Block Num = Ceil(输入token数/cache_block_size)+Celi(平均输出token数/cache_block_size)
    """
    if any(arg <= 0 for arg in locals().values()):
        return 0, 0, 0
    
    input_block_num = math.ceil(input_tokens / cache_block_size)
    output_block_num = math.ceil(output_tokens / cache_block_size)
    avg_block_num = math.ceil(avg_tokens / cache_block_size)

    return (
        input_block_num + output_block_num, # max
        input_block_num, # min
        input_block_num + avg_block_num # avg
    )


def cal_max_batch_size_range(
    total_block_num: int,
    max_block: int,
    min_block: int,
    avg_block: int,
    sp: int
):
    """
        最小maxBatchSize = Floor[Total Block Num/(所需最大Block Num/SP)]
        最大maxBatchSize = Floor[Total Block Num/(所需最小Block Num/SP)]
        平均maxBatchSize = Floor[Total Block Num/(所需平均Block Num/SP)]
    """
    if any(arg <= 0 for arg in locals().values()):
        return 0, 0, 0
    
    logger.debug(f"sp: {sp}")
    return (
        total_block_num // math.ceil(max_block / sp), # min
        total_block_num // math.ceil(min_block / sp), # max
        total_block_num // math.ceil(avg_block / sp)  # avg
    )


def write_to_answer(min_batch, max_batch, avg_batch):
    if any(arg <= 0 for arg in locals().values()):
        return
    
    answer(
        suggesion_type=SUGGESTION_TYPES.config,
        suggesion_item="maxBatchSize",
        action=f"set to range [{min_batch}, {max_batch}], average is {avg_batch}",
        reason="经过对当前显存信息的计算，建议将maxBatchSize的值设置为average大小，并逐渐向范围最大值调整，以占满整个显存",
    )


@register_analyze()
def find_max_batch_size_range(server_config, benchmark, output_log, input_params):
    # get input token number and average output token num
    input_token_num, avg_token_num = extract_token_num(benchmark, input_params)

    # get server info from server config.json
    try:
        server_params = extract_server_config_params(server_config, input_params)
    except Exception as e:
        logger.warning(f"Skip npu memory analyze due to {e}, please check mindie-server config permission or content.")
        return
    
    # get model info from model config.json
    try:
        model_params, model_weight_size = extract_model_config_params(server_params["model_weight_path"])
    except Exception as e:
        logger.warning(f"Skip npu memory analyze due to invaild model weight. {e}")
        return

    # caculate npu memory
    try:
        npu_mem_size = cal_npu_mem_size(server_params, model_weight_size)
        total_block_num = cal_total_block_num(npu_mem_size, server_params, model_params)
        logger.info(f"total_block_num:{total_block_num}")
    except Exception as e:
        logger.warning(f"Skip npu memory analyze due to npu memory calculation failed. {e}")
        return
    
    # caculate block numbers
    max_block, min_block, avg_block = cal_block_nums(
        input_token_num, server_params['cache_block_sizes'], server_params['output_token_num'], avg_token_num)
    logger.info(f"max_block_num:{max_block}  min_block_num:{min_block}  avg_block_num:{avg_block}")

    # caculate max_batch_size
    min_batch, max_batch, avg_batch = cal_max_batch_size_range(
        total_block_num, max_block, min_block, avg_block, server_params['sp'])
    logger.info(f"max_batch_size:{max_batch}  min_batch_size:{min_batch}  avg_batch_size:{avg_batch}")

    write_to_answer(min_batch, max_batch, avg_batch)



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

import numpy as np
import pandas as pd
from ms_service_profiler.exporters.base import ExporterBase
from ms_service_profiler.utils.log import logger

from ..common.csv_fields import RequestCSVFields, BatchCSVFields, ServiceCSVFields


def is_contained_valid_iter_info(rid_list, token_id_list):
    if rid_list is None or token_id_list is None or len(rid_list) != len(token_id_list):
        return False
    return True


def print_warning_log(log_name):
    if not ExporterSummary.get_err_log_flag(log_name):
        logger.warning(f"The '{log_name}' field info is missing, please check.")
        ExporterSummary.set_err_log_flag(log_name, True)


def process_each_record(req_map, batch_map, record):
    process_req_record(req_map, record)
    process_batch_record(batch_map, record)


def process_batch_record(batch_map, record):
    batch_type = record.get('batch_type')
    rid_tuple = str(record.get('rid_list'))

    # 构建batch_map
    if batch_type == 'Prefill':
        prefill_key = f"prefill_{rid_tuple}"
        if prefill_key not in batch_map:
            batch_map[prefill_key] = {
                BatchCSVFields.PREFILL_BATCH_NUM: 0,
                BatchCSVFields.PREFILL_EXEC_TIME: 0.0
            }
        batch_map[prefill_key][BatchCSVFields.PREFILL_BATCH_NUM] = int(record.get('batch_size'))
        batch_map[prefill_key][BatchCSVFields.PREFILL_EXEC_TIME] += float(record.get('during_time')) / 1000

    if batch_type == 'Decode':
        decode_key = f"decode_{rid_tuple}"
        if decode_key not in batch_map:
            batch_map[decode_key] = {
                BatchCSVFields.DECODE_BATCH_NUM: 0,
                BatchCSVFields.DECODE_EXEC_TIME: 0.0
            }
        batch_map[decode_key][BatchCSVFields.DECODE_BATCH_NUM] = int(record.get('batch_size'))
        batch_map[decode_key][BatchCSVFields.DECODE_EXEC_TIME] += float(record.get('during_time')) / 1000


def process_req_record(req_map, record):
    name = record.get('name')
    rid = record.get('rid')
    req_wait_status = record.get('WAITING+')
    req_pend_status = record.get('PENDING+')
    recv_token = record.get('recvTokenSize=')
    reply_token = record.get('replyTokenSize=')

    if rid is None or name is None:
        print_warning_log('rid or name')
        return

    # req_map中存在rid则保存，不存在则默认
    if rid not in req_map:
        req_map[rid] = {
            'httpReq_start': None,
            'httpRes_end': None,
            'token_id': {},
            'req_waiting_time': 0.0,
            'req_pending_time': 0.0,
            'is_complete': False,
            'generated_token_num': None,
            'input_token_num': None,
            'exec_time': 0.0,
            'first_token_latency': 0.0,
        }
    entry = req_map[rid]

    # 处理不同name的记录
    if name == 'httpReq':
        entry['httpReq_start'] = record.get('start_time')
        entry['token_id'] = entry.get('token_id', {})  # 确保token_id存在
        entry.setdefault('req_waiting_time', 0.0)
        entry.setdefault('req_pending_time', 0.0)
        entry['is_complete'] = False  # 重置完成状态
    elif name == 'httpRes':
        entry['httpRes_end'] = record.get('end_time')
        entry['is_complete'] = True

    # 处理队列状态时间
    if req_wait_status == 1:
        entry['req_waiting_time'] = record.get('during_time', 0.0)
    if req_pend_status == 1:
        entry['req_pending_time'] = record.get('during_time', 0.0)

    # 处理Token数量
    if recv_token is not None:
        entry['input_token_num'] = recv_token
    if reply_token is not None:
        entry['generated_token_num'] = reply_token

    # 处理rid_list和token_id_list
    rid_list = record.get('rid_list')
    token_id_list = record.get('token_id_list')
    process_rid_token_list(req_map, rid_list, token_id_list, record)

    # 检查并移除空数据
    if is_empty_entry(entry):
        req_map.pop(rid)


def is_empty_entry(entry):
    """
    检查entry是否为空数据，即为默认值的场景
    """
    return (
        entry['httpReq_start'] is None
        and entry['httpRes_end'] is None
        and not entry['token_id']  # 空字典
        and entry['req_waiting_time'] == 0.0
        and entry['req_pending_time'] == 0.0
        and entry['is_complete'] is False
        and entry['generated_token_num'] is None
        and entry['input_token_num'] is None
    )


def process_rid_token_list(req_map, rid_list, token_id_list, record):
    if not is_contained_valid_iter_info(rid_list, token_id_list):
        return

    for i, value in enumerate(rid_list):
        req_rid = str(value)
        if req_map.get(req_rid) is None:
            print_warning_log('httpReq')
            continue

        # 执行总时长
        if record.get('name') == 'modelExec':
            during_time = record.get('during_time')
            if isinstance(during_time, (int, float)):
                req_map[req_rid]['exec_time'] += during_time
            else:
                logger.warning(f"Invalid during_time: {during_time} for rid={req_rid}")

        cur_iter = token_id_list[i]
        if cur_iter is None:
            print_warning_log('token_id_list')
            continue

        req_map[req_rid]['token_id'][str(cur_iter)] = record.get('end_time')

        # 首Token时延
        if cur_iter == 0:
            if req_map[req_rid].get('first_token_latency') is None:
                req_map[req_rid]['first_token_latency'] = record.get('during_time')
            else:
                req_map[req_rid]['first_token_latency'] += record.get('during_time')


def gen_exporter_results(all_data_df):
    req_map = {}
    batch_map = {}

    for _, record in all_data_df.iterrows():
        process_each_record(req_map, batch_map, record)

    # 生成request维度数据
    req_view, total_map = calculate_request_metrics(req_map)

    # 获取req_view数据
    first_token_latency = [req["first_token_latency"] for req in req_view]
    subsequent_token_latency = [
        latency
        for req in req_view
        for latency in req["subsequent_token_latency"]
    ]
    total_time = [req["total_time"] for req in req_view]
    exec_time = [req["exec_time"] for req in req_view]
    waiting_time = [req["waiting_time"] for req in req_view]
    input_token_num = [req["input_token_num"] for req in req_view]
    generated_token_num = [req["generated_token_num"] for req in req_view]

    # 计算统计值
    req_stats = {
        RequestCSVFields.FIRST_TOKEN_LATENCY: calculate_statistics(first_token_latency),
        RequestCSVFields.SUBSEQUENT_TOKEN_LATENCY: calculate_statistics(subsequent_token_latency),
        RequestCSVFields.TOTAL_TIME: calculate_statistics(total_time),
        RequestCSVFields.EXEC_TIME: calculate_statistics(exec_time),
        RequestCSVFields.WAITING_TIME: calculate_statistics(waiting_time),
        RequestCSVFields.INPUT_TOKEN_NUM: calculate_statistics(input_token_num),
        RequestCSVFields.GENERATED_TOKEN_NUM: calculate_statistics(generated_token_num)
    }

    # 生成batch维度数据
    batch_status = calculate_batch_metrics(batch_map)

    return req_stats, batch_status, total_map


def calculate_statistics(metric):
    """
    计算avg、max、min、p99、p90、p50的公共函数
    """
    if not metric or any(not isinstance(value, (int, float)) for value in metric):
        return {
            "avg": np.nan,
            "max": np.nan,
            "min": np.nan,
            "p50": np.nan,
            "p90": np.nan,
            "p99": np.nan
        }
    return {
        "avg": round(np.mean(metric), 4),
        "max": round(np.max(metric), 4),
        "min": round(np.min(metric), 4),
        "p50": round(np.percentile(metric, 50), 4),
        "p90": round(np.percentile(metric, 90), 4),
        "p99": round(np.percentile(metric, 99), 4)
    }


def calculate_batch_metrics(batch_map):
    prefill_batch_num_list = []
    decode_batch_num_list = []
    prefill_exec_time_list = []
    decode_exec_time_list = []

    for key, value in batch_map.items():
        if key.startswith('prefill'):
            prefill_batch_num_list.append(value.get(BatchCSVFields.PREFILL_BATCH_NUM, 0))
            prefill_exec_time_list.append(value.get(BatchCSVFields.PREFILL_EXEC_TIME, 0.0))
        elif key.startswith('decode'):
            decode_batch_num_list.append(value.get(BatchCSVFields.DECODE_BATCH_NUM, 0))
            decode_exec_time_list.append(value.get(BatchCSVFields.DECODE_EXEC_TIME, 0.0))

    # 计算统计指标
    batch_status = {
        BatchCSVFields.PREFILL_BATCH_NUM: calculate_statistics(prefill_batch_num_list),
        BatchCSVFields.DECODE_BATCH_NUM: calculate_statistics(decode_batch_num_list),
        BatchCSVFields.PREFILL_EXEC_TIME: calculate_statistics(prefill_exec_time_list),
        BatchCSVFields.DECODE_EXEC_TIME: calculate_statistics(decode_exec_time_list)
    }

    return batch_status


def get_non_first_token_latency(req_data):
    token_id = req_data["token_id"]
    subsequent_token_latency = []
    sorted_tokens = sorted(token_id.items(), key=lambda x: int(x[0]))
    for i in range(1, len(sorted_tokens)):
        current_token_time = sorted_tokens[i][1]
        previous_token_time = sorted_tokens[i - 1][1]
        latency = round((current_token_time - previous_token_time) / 1000, 4)
        subsequent_token_latency.append(latency)
    return subsequent_token_latency


def gen_result_record(req_id, req_data):
    input_token_num = req_data["input_token_num"]
    generated_token_num = req_data["generated_token_num"]

    # 从字典中取出数据
    record = {
        "req_id": req_id,
        "first_token_latency": round(req_data["first_token_latency"] / 1000, 4),
        "exec_time": round(req_data["exec_time"] / 1000, 4),
        "input_token_num": input_token_num,
        "generated_token_num": generated_token_num
    }

    # 检查 httpReq_start 和 httpRes_end 的有效性
    if req_data["httpReq_start"] is not None and req_data["httpRes_end"] is not None:
        total_time = (req_data["httpRes_end"] - req_data["httpReq_start"]) / 1000
        record["total_time"] = round(total_time, 4)
    else:
        record["total_time"] = 0.0  # 无效数据

    # 非首Token时延
    record["subsequent_token_latency"] = get_non_first_token_latency(req_data)

    # 队列等待时长
    waiting_time = req_data["req_pending_time"] + req_data["req_waiting_time"]
    record["waiting_time"] = waiting_time
    return record


def calculate_request_metrics(req_map):
    req_view = []
    total_map = {
        ServiceCSVFields.TOTAL_INPUT_TOKEN_NUM: 0,
        ServiceCSVFields.TOTAL_GENERATED_TOKEN_NUM: 0,
        ServiceCSVFields.GENERATE_TOKEN_SPEED: 0,
        ServiceCSVFields.GENERATE_ALL_TOKEN_SPEED: 0
    }

    valid_requests = []
    first_request_start_time, last_request_end_time = None, None

    # 过滤无效请求时同时保存 req_id
    for req_id, req_data in req_map.items():
        if req_data["httpReq_start"] is not None and req_data["httpRes_end"] is not None:
            valid_requests.append((req_id, req_data))

    # 处理有效请求时解包 req_id 和 req_data
    for req_id, req_data in valid_requests:
        record = gen_result_record(req_id, req_data)
        req_view.append(record)

        # 更新总体维度数据
        total_map[ServiceCSVFields.TOTAL_INPUT_TOKEN_NUM] += req_data["input_token_num"]
        total_map[ServiceCSVFields.TOTAL_GENERATED_TOKEN_NUM] += req_data["generated_token_num"]

        # 更新第一个和最后一个请求的时间
        current_start_time = req_data["httpReq_start"]
        current_end_time = req_data["httpRes_end"]
        if first_request_start_time is None or current_start_time < first_request_start_time:
            first_request_start_time = current_start_time
        if last_request_end_time is None or current_end_time > last_request_end_time:
            last_request_end_time = current_end_time

    total_exec_time = 0.0

    # 计算 total_exec_time
    if first_request_start_time is not None and last_request_end_time is not None:
        total_exec_time = round((last_request_end_time - first_request_start_time) / 1000000, 4)

    if total_exec_time > 0:
        total_map[ServiceCSVFields.GENERATE_TOKEN_SPEED] = round(
            total_map[ServiceCSVFields.TOTAL_GENERATED_TOKEN_NUM] / total_exec_time, 4
        )
        total_map[ServiceCSVFields.GENERATE_ALL_TOKEN_SPEED] = round(
            (total_map[ServiceCSVFields.TOTAL_INPUT_TOKEN_NUM] +
             total_map[ServiceCSVFields.TOTAL_GENERATED_TOKEN_NUM]) / total_exec_time, 4
        )

    return req_view, total_map


def save_dataframe_to_csv(map_data, output, file_name, include_stats=1):
    if output is None:
        return

    output_path = Path(output)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    file_path = output_path / file_name

    # 将map_data转换为DataFrame
    df = convert_map_to_dataframe(map_data, include_stats)

    # 保存到csv
    df.to_csv(file_path, index=False)

    os.chmod(file_path, 0o640)


def convert_map_to_dataframe(map_data, include_stats):
    data = []
    for metric, values in map_data.items():
        if include_stats == 1:
            row = {
                BatchCSVFields.METRIC: metric,
                BatchCSVFields.AVG: values["avg"],
                BatchCSVFields.MAX: values["max"],
                BatchCSVFields.MIN: values["min"],
                BatchCSVFields.P50: values["p50"],
                BatchCSVFields.P90: values["p90"],
                BatchCSVFields.P99: values["p99"]
            }
        else:
            value = format(values, ".8f") if metric == ServiceCSVFields.GENERATE_TOKEN_SPEED else values
            row = {ServiceCSVFields.METRIC: metric, ServiceCSVFields.VALUE: value}
        data.append(row)
    return pd.DataFrame(data)


class ExporterSummary(ExporterBase):
    name = "summary"
    err_log = {'rid or name': False, 'start_time': False, 'httpReq': False, 'token_id_list': False}

    @classmethod
    def initialize(cls, args):
        cls.args = args
        cls.err_log = {'rid or name': False, 'start_time': False, 'httpReq': False, 'token_id_list': False}

    @classmethod
    def set_err_log_flag(cls, index, value):
        cls.err_log[index] = value

    @classmethod
    def get_err_log_flag(cls, index):
        return cls.err_log[index]

    @classmethod
    def export(cls, data) -> None:
        all_data_df = data.get('tx_data_df')

        if all_data_df is None:
            logger.warning("The data is empty, please check")
            return

        # 调用计算首Token时延的函数
        req_status, batch_status, total_map = gen_exporter_results(all_data_df)
        output = cls.args.output_path

        # 格式化存入csv
        save_dataframe_to_csv(req_status, output, RequestCSVFields.PATH_NAME)
        save_dataframe_to_csv(batch_status, output, BatchCSVFields.PATH_NAME)
        save_dataframe_to_csv(total_map, output, ServiceCSVFields.PATH_NAME, include_stats=0)




# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
from ms_service_profiler.utils.log import logger
from .constants import MAX_BATCH_NUMBER


EVENT_PAIRS = [
    ('batchFrameworkProcessing', 'preprocessBatch'),
    ('preprocessBatch', 'serializeExcueteMessage'),
    ('serializeExcueteMessage', 'deserializeExecuteRequestsForInfer'),
    ('grpcWriteToSlave', 'deserializeExecuteRequestsForInfer'),
    ('deserializeExecuteRequestsForInfer', 'convertTensorBatchToBackend'),
    ('convertTensorBatchToBackend', 'getInputMetadata'),
    ('getInputMetadata', 'preprocess'),
    ('getInputMetadata', 'prepareInputs'),
    ('forward', 'sample'),
    ('operatorExecute', 'sample'),
    ('postprocess', 'generateOutput'),
    ('generateOutput', 'processPythonExecResult'),
    ('deserializeExecuteResponse', 'saveoutAndContinueBatching'),
]

SYNC_EVENT_PAIRS = [
    ('processBroadcastMessage', 'deserializeExecuteRequestsForInfer'),
    ('getInputMetadata', 'preprocess'),
    ('getInputMetadata', 'prepareInputs'),
    ('preprocess', 'forward'),
    ('forward', 'sample'),
    ('prepareInputs', 'operatorExecute'),
    ('operatorExecute', 'sample'),
    ('getInputMetadata', 'sample'),
    ('sample', 'postprocess'),
    ('postprocess', 'processPythonExecResult'),
]

FILTER_LIST = ['deserializeExecuteRequestsForInfer', 'deserializeExecuteResponse', 'convertTensorBatchToBackend', 
               'processPythonExecResult', 'forward', 'sample', 'setInferBuffer', 'grpcWriteToSlave']

NAME_LIST_MINDIE = ['httpReq', 'encode', 'preprocessBatch', 'batchFrameworkProcessing', 'serializeExcueteMessage',
             'deserializeExecuteRequestsForInfer', 'convertTensorBatchToBackend', 'getInputMetadata',
             'preprocess', 'forward', 'sample', 'postprocess', 'generateOutput', 'prepareInputs', 'operatorExecute',
             'processPythonExecResult', 'deserializeExecuteResponse', 'saveoutAndContinueBatching', 'continueBatching',
             'setInferBuffer', 'grpcWriteToSlave', 'receiveInfer', 'handleTaskExecution', 'decodeEnd', 'httpRes']

NAME_LIST_VLLM = ['httpReq', 'encode', 'batchFrameworkProcessing', 'preprocess', 'forward', 'httpRes']

FULL_BATCH = ['serializeExcueteMessage', 'deserializeExecuteRequestsForInfer', 'convertTensorBatchToBackend', 
              'getInputMetadata', 'preprocess', 'forward', 'sample', 'postprocess']

CSV_COLUMNS = ['name', 'during_time', 'pid', 'tid', 'start_time', 'end_time', 'rid', 'start_datetime', 
               'end_datetime', 'batch_type', 'batch_size', 'rid_list', 'token_id_list']

HTTP_LIST = ['encode', 'httpReq', 'decodeEnd', 'httpRes']

RENAMED_COLUMNS = {
        'start_time': 'start_time(microsecond)',
        'end_time': 'end_time(microsecond)',
        'during_time': 'during_time(microsecond)'
    }


def get_name_list(service_type):
    if service_type == 'mindie':
        name_list = NAME_LIST_MINDIE
    else:
        name_list = NAME_LIST_VLLM

    return name_list



def preprocess_framework_df(framework_df):
    try:
        framework_df = framework_df[framework_df['name'].isin(NAME_LIST_MINDIE)]
        framework_df = framework_df[CSV_COLUMNS]
        framework_df = framework_df.rename(columns=RENAMED_COLUMNS)
    except KeyError as e:
        logger.warning(f"Field '{e.args[0]}' not found in msproftx.db.")
        return None

    return framework_df


def is_valid_prefill(batch_group, rid, framework_df):
    batch_row = batch_group.iloc[0]
    cur_rid = batch_row['rid_list'][0]
    if rid != '-1':
        cur_rid = rid

    target_encode = framework_df[(framework_df['rid'] == str(cur_rid)) & 
                                 (framework_df['name'] == 'httpReq')]
    return not target_encode.empty


def get_pre_half_batch(full_batch, index_list, framework_df, pid, tid):
    df_list = []
    for name, i in zip(full_batch, index_list):
        try:
            if name == 'serializeExcueteMessage':
                index = i
            elif name == 'deserializeExecuteRequestsForInfer' or name == 'convertTensorBatchToBackend':
                index = framework_df[(framework_df.index >= i) &
                            (framework_df['name'] == name) &
                            (framework_df['pid'] == pid)].index[0]
            else:
                index = framework_df[(framework_df.index >= i) &
                            (framework_df['name'] == name) &
                            (framework_df['pid'] == pid) &
                            (framework_df['tid'] == tid)].index[0]
        except IndexError:
            logger.warning(f"no named {name} line, skip this batch")
            return pd.DataFrame()
        df_list.append(index)
    return pd.concat([framework_df.loc[df_list]])
    

def get_last_half_batch(framework_df, generate_index, pid, tid):
    result = pd.DataFrame()
    index = []
    try:
        py_result_index = framework_df[(framework_df.index >= generate_index) &
                            (framework_df['name'] == 'processPythonExecResult') &
                            (framework_df['pid'] == pid) & (framework_df['tid'] == tid)].index[0]
        deserialize_index = framework_df[(framework_df.index > py_result_index) &
                            (framework_df['name'] == 'deserializeExecuteResponse')].index[0]
    except IndexError:
        logger.warning(f"no match line, skip this batch")
        return result
    index.extend([py_result_index, deserialize_index])
    result = pd.concat([result, framework_df.loc[index]])
    return result


def get_full_batch_vllm(group, framework_df):
    result = pd.DataFrame()
    try:
        batch_start_index = group[group['name'] == 'batchFrameworkProcessing'].index[0]
        all_time_index = group[group['name'] == 'AllTime'].index[0]
        concat_list = [batch_start_index, all_time_index]
        preprocess_row = framework_df[(framework_df.index > batch_start_index) &
                                    (framework_df['name'] == 'preprocess')].iloc[0]
        preprocess_pid, preprocess_tid = preprocess_row['pid'], preprocess_row['tid']
        preprocess_index = framework_df[(framework_df.index > batch_start_index) &
                                      (framework_df['name'] == 'preprocess')].index[0]
        forward_index = framework_df[(framework_df.index > batch_start_index) &
                                      (framework_df['name'] == 'forward') &
                                      (framework_df['pid'] == preprocess_pid) &
                                      (framework_df['tid'] == preprocess_tid)].index[0]
    except IndexError:
        logger.warning("no match line, skip this batch")
        return result
    concat_list.extend([preprocess_index, forward_index])
    result = pd.concat([result, framework_df.loc[concat_list]])
    result = result.sort_values(by='start_time(microsecond)').reset_index(drop=True)
    return result


def get_full_batch(group, framework_df):
    service_type = get_service_type(framework_df)
    if service_type == 'vllm':
        return get_full_batch_vllm(group, framework_df)

    name_list = get_name_list(service_type)
    start_index = name_list.index('batchFrameworkProcessing')
    end_index = name_list.index('saveoutAndContinueBatching')
    batch_start_index = group[group['name'] == 'batchFrameworkProcessing'].index[0]
    concat_list = [batch_start_index]
    index_list = []
    index = batch_start_index
    # 找到generator_row
    for name in FULL_BATCH:
        query = f"index > @index and name =='{name}'"
        if framework_df.query(query).empty:
            logger.warning(f"no named {name} line, skip this batch")
            return pd.DataFrame()
        index = framework_df[(framework_df.index > index) &
                             (framework_df['name'] == name)].index[0]
        index_list.append(index)
    if framework_df.query('name == "generateOutput" and index > @index').empty:
        logger.warning(f"no generateOutput line, skip this batch")
        return pd.DataFrame()
    generate_row = framework_df[(framework_df.index > index) &
                                (framework_df['name'] == 'generateOutput')].iloc[0]
    generate_pid, generate_tid = generate_row['pid'], generate_row['tid']
    generate_index = framework_df[(framework_df.index > index) &
                                  (framework_df['name'] == 'generateOutput')].index[0]
    concat_list.append(generate_index)

    # 找到前半部分的字段
    result_pre = get_pre_half_batch(FULL_BATCH, index_list, framework_df, generate_pid, generate_tid)
    # 拼接后半部分字段
    result_last = get_last_half_batch(framework_df, generate_index, generate_pid, generate_tid)
    if result_last.empty or result_pre.empty:
        return pd.DataFrame()
    group = group.drop(group[group['name'].isin(name_list[start_index: end_index])].index)
    result = pd.concat([framework_df.loc[concat_list], result_pre, result_last, group])
    result = result.sort_values(by='start_time(microsecond)').reset_index(drop=True)
    return result


def get_groups(framework_df, batch_size, rid, name):
    result_df = []
    
    groups = framework_df.groupby((framework_df['name'] == 'batchFrameworkProcessing').cumsum())
    for _, group in groups:
        if rid != '-1':
            batch_group = group[(group['name'] == 'batchFrameworkProcessing') &
                                (group['batch_type'] == name)]
            batch_group.loc[:, 'rid_list'] = batch_group['rid_list'].apply(lambda x: [str(i) for i in x])
            batch_group = batch_group[batch_group['rid_list'].apply(lambda x: rid in x)]
        elif batch_size > 0:
            batch_group = group[(group['name'] == 'batchFrameworkProcessing') &
                                (group['batch_type'] == name) &
                                (group['batch_size'] == str(batch_size))]
        if batch_group.empty:
            continue

        if name == 'Prefill' and not is_valid_prefill(batch_group, rid, framework_df):
            continue
        result = get_full_batch(group, framework_df)

        if not result.empty:
            result_df.append(result) 

    return result_df


def get_event_pair_df(framework_df, name):
    new_rows = []
    # 遍历每个事件对
    for current_name, next_name in EVENT_PAIRS:
        # 收集当前事件的所有实例
        current_events = framework_df[framework_df['name'] == current_name].copy()
        len_current = len(current_events)
        if len_current == 0:
            logger.debug(f'{name}: No data named {current_name}')
            continue
        next_events = framework_df[framework_df['name'] == next_name].copy()
        len_next = len(next_events)
        if len_next == 0:
            logger.debug(f'{name}: No data named {next_name}')
            continue
        # 提取当前事件的pid，tid
        for i in range(len_current):
            cur_pid = current_events.iloc[i]['pid']
            cur_tid = current_events.iloc[i]['tid']
            if (current_name, next_name) in SYNC_EVENT_PAIRS:
                next_events_filter = next_events[(next_events['pid'] == cur_pid) & (next_events['tid'] == cur_tid)]
                if len(next_events_filter) == 0:
                    continue
                current_end = current_events.iloc[i]['end_time(microsecond)']
                next_start = next_events_filter.iloc[0]['start_time(microsecond)']
                if next_start < current_end:
                    continue
            else:
                next_events_filter = next_events[next_events['pid'] == cur_pid]
                if len(next_events_filter) == 0:
                    continue
                current_end = current_events.iloc[i]['end_time(microsecond)']
                next_start = next_events_filter.iloc[0]['start_time(microsecond)']
                if next_start < current_end:
                    continue
            # 创建新行
            new_row = {
                'name': f"Between-{current_name}-{next_name}",
                'during_time(microsecond)': next_start - current_end,
                'start_time(microsecond)': current_end,
                'end_time(microsecond)': next_start,
                'pid': current_events.iloc[i]['pid'],
                'tid': current_events.iloc[i]['tid'],
            }
            new_rows.append(new_row)

    # 将新行转换为 DataFrame
    new_df = pd.DataFrame(new_rows)

    # 合并原数据和新行，并按时间排序
    result_df = pd.concat([framework_df, new_df], ignore_index=True)
    result_df = result_df.sort_values(by='start_time(microsecond)').reset_index(drop=True)
    return result_df


def get_post_event_df(post_event_pairs, framework_df, name):
    new_rows = []
    # 遍历每个事件对
    for current_name, next_name in post_event_pairs:
        # 收集当前事件的所有实例
        current_events = framework_df[framework_df['name'] == current_name].copy()
        len_current = len(current_events)
        if len_current == 0:
            logger.debug(f'{name}: No data named {current_name}')
            continue
        next_events = framework_df[framework_df['name'] == next_name].copy()
        len_next = len(next_events)
        if len_next == 0:
            logger.debug(f'{name}: No data named {next_name}')
            continue
        current_end = current_events.iloc[0]['end_time(microsecond)']
        next_start = next_events.iloc[0]['start_time(microsecond)']
        # 创建新行
        new_row = {
            'name': f"Between-{current_name}-{next_name}",
            'during_time(microsecond)': next_start - current_end,
            'start_time(microsecond)': current_end,
            'end_time(microsecond)': next_start,
            'pid': current_events.iloc[0]['pid'],
            'tid': current_events.iloc[0]['tid'],
        }
        new_rows.append(new_row)
    new_df = pd.DataFrame(new_rows)

    return new_df


def postprocess_framework_df(framework_df, post_event_pairs, name):
    if framework_df.empty:
        logger.warning(f'{name}: df is empty')
        return framework_df

    service_type = get_service_type(framework_df)
    
    # 更新AllTime行
    framework_df.loc[framework_df['name'] == 'AllTime', ['start_time(microsecond)', 'end_time(microsecond)']] = \
        framework_df.loc[framework_df['name'] == 'AllTime', ['end_time(microsecond)', 'start_time(microsecond)']].values

    if service_type == 'mindie':
        if name == 'Prefill':
            if 'continueBatching' not in framework_df['name'].values:
                post_event_pairs.append(('deserializeExecuteResponse', 'httpRes'))
                if 'httpRes' not in framework_df['name'].values:
                    post_event_pairs.append(('deserializeExecuteResponse', 'AllTime'))
        new_rows = []
        if 'preprocessBatch' not in framework_df['name'].values:
            post_event_pairs.append(('batchFrameworkProcessing', 'serializeExcueteMessage'))
        if 'continueBatching' not in framework_df['name'].values and name == 'Decode':
            post_event_pairs.append(('deserializeExecuteResponse', 'AllTime'))
    elif service_type == 'vllm':
        start_index = NAME_LIST_VLLM.index('batchFrameworkProcessing')
        end_index = NAME_LIST_VLLM.index('httpRes')
        key_names_vllm = NAME_LIST_VLLM[start_index:end_index]
        key_event_pairs = [(key_names_vllm[i], key_names_vllm[i + 1]) for i in range(len(key_names_vllm) - 1)]
        for pair in key_event_pairs:
            post_event_pairs.append(pair)
        post_event_pairs.append(('forward', 'AllTime'))


    
    post_df = get_post_event_df(post_event_pairs, framework_df, name)
    framework_df = pd.concat([framework_df, post_df], ignore_index=True)
    framework_df = framework_df.sort_values(by=['start_time(microsecond)']).reset_index(drop=True)
    framework_df = framework_df[framework_df['name'] != 'handleTaskExecution']  # 删除 'handleTaskExecution' 行
    if name == 'Decode':
        framework_df = framework_df.drop(framework_df[framework_df['name'].isin(HTTP_LIST)].index)
    framework_df['during_time(microsecond)'] = framework_df['during_time(microsecond)'] / 1000
    
    if 'prepareInputs' in framework_df['name'].values and 'operatorExecute' in framework_df['name'].values:
        delete_list = ['preprocess', 'forward', 'Between-getInputMetadata-preprocess', 
                       'Between-preprocess-forward', 'Between-forward-sample']
    else:
        delete_list = ['prepareInputs', 'operatorExecute', 'Between-getInputMetadata-prepareInputs', 
                       'Between-prepareInputs-operatorExecute', 'Between-operatorExecute-sample']

    framework_df = framework_df[~framework_df['name'].isin(delete_list)]
    all_time_rows = framework_df[framework_df['name'] == 'AllTime']
    framework_df = framework_df[framework_df['name'] != 'AllTime']
    framework_df = pd.concat([framework_df, all_time_rows], ignore_index=True)
    framework_df.drop_duplicates(subset='name', inplace=True)
    return framework_df


def get_filter_rule_df(framework_df):
    serialize_execute_index = framework_df[framework_df['name'] == 'serializeExcueteMessage'].index
    if len(serialize_execute_index) != 0:
        serialize_execute_index = serialize_execute_index[0]
        rows_to_drop = framework_df[(framework_df.index < serialize_execute_index) &
                                    (framework_df['name'].isin(FILTER_LIST))].index
        framework_df = framework_df.drop(rows_to_drop)
    
    # 删除 'name' 为 'encode'、'httpReq'、'decode'、'httpRes' 的行
    framework_df = framework_df.drop(framework_df[framework_df['name'].isin(HTTP_LIST)].index)
    return framework_df


def get_batch_framework(framework_df, name):
    if name == 'Prefill':
        framework_df = get_filter_rule_df(framework_df)
    framework_df = get_event_pair_df(framework_df, name)

    grouped_df = framework_df.groupby('name')
    sorted_groups = grouped_df.apply(lambda x: x.sort_values(by='during_time(microsecond)', ascending=True))
    if 'generateOutput' in grouped_df.groups:
        generate_output_pid = grouped_df.get_group('generateOutput')['pid'].iloc[0]
        generate_output_tid = grouped_df.get_group('generateOutput')['tid'].iloc[0]
    elif 'forward' in grouped_df.groups:
        sample_group = grouped_df.get_group('forward')
        last_sample_row = sample_group.tail(1)
        generate_output_pid = last_sample_row['pid'].values[0]
        generate_output_tid = last_sample_row['tid'].values[0]
    else:
        return pd.DataFrame()

    filter_name = ['deserializeExecuteRequestsForInfer', 'convertTensorBatchToBackend',
                   'Between-deserializeExecuteRequestsForInfer-convertTensorBatchToBackend',
                   'Between-convertTensorBatchToBackend-getInputMetadata']
    result_df = sorted_groups[(sorted_groups['pid'] == generate_output_pid) &
                              (sorted_groups['tid'] == generate_output_tid)]
    result_df1 = sorted_groups[(sorted_groups['pid'] == generate_output_pid) &
                               (sorted_groups['name'].isin(filter_name))]
    result_df = pd.concat([result_df, result_df1])
    other_group = sorted_groups[sorted_groups['pid'] != generate_output_pid]
    other_group = other_group[~other_group['name'].isin(result_df['name'])]
    other_df = other_group.drop_duplicates(subset='name', keep='first')
    result_df = pd.concat([result_df, other_df])
    return result_df


def get_filter_df(framework_df, name):
    """
    从第一条httpReq开始
    """
    if name == 'Prefill':
        filter_name = 'httpReq'
    else:
        filter_name = 'batchFrameworkProcessing'
    first_batch_schedule_index = 0
    try:
        first_batch_schedule_index = framework_df[framework_df['name'] == filter_name].index[0]
    except IndexError:
        logger.warning(f"{name}: No data named {filter_name}")
    filter_df = framework_df.loc[first_batch_schedule_index:]
    return filter_df


def get_statistics_data(framework_df, filter_name, name):
    if framework_df.empty:
        logger.warning(f"{name}: The dataframe is empty, no csv file create")
        return framework_df
    batch = framework_df[framework_df['name'] == filter_name]
    if len(batch) == 1:
        start_index = framework_df[framework_df['name'] == filter_name].index[-1]
        end_index = None
    else:
        start_index = framework_df[framework_df['name'] == filter_name].index[-2]
        end_index = framework_df[framework_df['name'] == filter_name].index[-1] - 1
    framework_df['max'] = framework_df.groupby('name')['during_time(microsecond)'].transform('max')
    framework_df['min'] = framework_df.groupby('name')['during_time(microsecond)'].transform('min')
    framework_df['mean'] = framework_df.groupby('name')['during_time(microsecond)'].transform('mean')
    framework_df['std'] = framework_df.groupby('name')['during_time(microsecond)'].transform('std')
    # 标准差为0时显示0
    framework_df['std'] = framework_df['std'].fillna(0)
    framework_df.insert(2, 'max', framework_df.pop('max'))
    framework_df.insert(3, 'min', framework_df.pop('min'))
    framework_df.insert(4, 'mean', framework_df.pop('mean'))
    framework_df.insert(5, 'std', framework_df.pop('std'))
    framework_df = framework_df.rename(columns={'during_time(microsecond)': 'during_time(millisecond)'})
    if name == 'Decode':
        framework_df = framework_df.iloc[:, :10]
    else:
        framework_df = framework_df.iloc[:, :11]
    return framework_df[start_index: end_index]


def get_batch_all_time(framework_df, name):
    new_rows = []
    batch_rows = framework_df[framework_df['name'] == 'batchFrameworkProcessing']
    pids = batch_rows['pid'].unique()
    for pid in pids:
        current_events = framework_df[(framework_df['name'] == 'batchFrameworkProcessing') & 
                                    (framework_df['pid'] == pid)].copy()

        len_current = len(current_events)

        if len_current < 2:
            logger.warning(f"{name}: The length of batchFrameworkProcessing is less two")
            return framework_df
        
        # 补充插入AllTime行
        for i in range(len_current - 1):
            current_row = current_events.iloc[i]
            next_row = current_events.iloc[i + 1]
            during_time = next_row['start_time(microsecond)'] - current_row['start_time(microsecond)']

            all_time_row = {
                    'name': 'AllTime',
                    'start_time(microsecond)': current_row['start_time(microsecond)'] + 2,
                    'end_time(microsecond)': next_row['start_time(microsecond)'],
                    'during_time(microsecond)': during_time,
                    'start_datetime': current_row['start_datetime'],
                    'end_datetime': next_row['start_datetime'],
                }

            # 将 'AllTime' 行添加到新行列表
            new_rows.append(all_time_row)

    new_df = pd.DataFrame(new_rows)
    result_df = pd.concat([framework_df, new_df], ignore_index=True)
    result_df = result_df.sort_values(by=['start_time(microsecond)', 'name']).reset_index(drop=True)

    return result_df


def get_batch_concat_df(filter_df, framework_df, cacl_num, rid, name):
    concat_df = pd.DataFrame()
    empty_row = pd.DataFrame(index=[0])
    for i in range(cacl_num):
        cur_rid = filter_df[i].iloc[0]['rid_list'][0]
        if rid != '-1':
            cur_rid = rid
        cur_df = get_batch_framework(filter_df[i], name)
        if cur_df.equals(pd.DataFrame()):
            continue
        if name == 'Prefill':
            http_df = framework_df[(framework_df['rid'] == str(cur_rid)) & (framework_df['name'].isin(HTTP_LIST))]
            cur_df = pd.concat([cur_df, http_df], ignore_index=True)
            post_event_pairs = [
                ('encode', 'batchFrameworkProcessing'),
                ('continueBatching', 'httpRes'),
            ]
        else:
            post_event_pairs = [
                ('continueBatching', 'AllTime'),
            ]
        cur_df = cur_df.sort_values(by='start_time(microsecond)')
        cur_df = postprocess_framework_df(cur_df, post_event_pairs, name)

        if cur_df.equals(pd.DataFrame()):
            logger.debug(f'{name}: cur_df is empty, continue')
            continue

        concat_df = pd.concat([concat_df, empty_row, cur_df], ignore_index=True)
    return concat_df


def process_exporter(framework_df, batch_size, batch_num, rid, name):
    # 划分组
    result_df = get_groups(framework_df, batch_size, rid, name)
    len_result_df = len(result_df)

    if len(result_df) == 0:
        logger.warning(f'{name}: no batchFrameworkProcessing with batch_size {batch_size}')
        logger.warning(f'{name}: no batchFrameworkProcessing with rid {rid}')
        return pd.DataFrame()
    
    cacl_num = len_result_df if len_result_df <= batch_num else batch_num
    cacl_num = min(cacl_num, MAX_BATCH_NUMBER)
    concat_df = get_batch_concat_df(result_df, framework_df, cacl_num, rid, name)
    return concat_df


def get_service_type(framework_df):
    name_set = set(list(framework_df['name']))
    if 'deserializeExecuteResponse' in name_set:
        service_type = 'mindie'
    else:
        service_type = 'vllm'
    return service_type


# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import csv
import argparse
from collections import namedtuple
from glob import glob
from dataclasses import dataclass

from msserviceprofiler.msservice_advisor.profiling_analyze.utils import TARGETS, LOG_LEVELS, SUGGESTION_TYPES
from msserviceprofiler.msservice_advisor.profiling_analyze.utils import str_ignore_case, logger, set_log_level

# {"21559056a7ff44c88a891ecbb537c431": "0", ...}
REQ_TO_DATA_MAP_PATTERN = "req_to_data_map.json"

# FirstTokenTime,DecodeTime,LastDecodeTime,...
# 213.2031 ms,228.3775 ms,88.327 ms,...
# -> average, max, min, P75, P90, SLO_P90, P99, N
RESULT_PERF_PATTERN = "result_perf_*.csv"
PERF_METRICS = ["average", "max", "min", "P75", "P90", "SLO_P90", "P99", "N"]
PERF_METRICS_MAP = {str_ignore_case(ii): ii for ii in PERF_METRICS}

# ...,Concurrency,ModelName,lpct,Throughput,GenerateSpeed,...
# ...,50,DeepSeek-R1,0.9336 ms,2.789 req/s,...
RESULT_COMMON_PATTERN = "result_common_*.csv"

# {"7": {"input_len": 213, "output_len": 12, "prefill_bsz": 15, "decode_bsz": [20, ...],
#        "req_latency": 2348332643508911, "latency": [798.7475395202637, ...], "queue_latency": [445314, ...], ... }
RESULTS_PER_REQUEST_PATTERN = "results_per_request_*.json"

MIES_INSTALL_PATH = "MIES_INSTALL_PATH"
MINDIE_SERVICE_DEFAULT_PATH = "/usr/local/Ascend/mindie/latest/mindie-service"

TARGETS_MAP = {
    "ttft": TARGETS.FirstTokenTime,
    "firsttokentime": TARGETS.FirstTokenTime,
    "throughput": TARGETS.Throughput,
}

LOG_LEVELS_LOWER = [ii.lower() for ii in LOG_LEVELS.keys()]


@dataclass
class ProfilingParameters:
    target: str
    target_metrics: str
    input_token_num: int
    output_token_num: int
    tp: int

    @classmethod
    def extract_from_args(cls, args):
        return cls(
            target=args.target,
            target_metrics=args.target_metrics,
            input_token_num=args.input_token_num,
            output_token_num=args.output_token_num,
            tp=args.tp
        )


""" parse_benchmark_instance """


def get_latest_matching_file(instance_path, pattern):
    files = glob(os.path.join(instance_path, pattern))
    return max(files, key=os.path.getmtime) if files else None


def read_csv(file_path):
    result = {}
    with open(file_path, mode="r", newline="", encoding="utf-8") as ff:
        for row in csv.DictReader(ff):
            for kk, vv in row.items():
                result.setdefault(kk, []).append(vv)
    return result


def read_json(file_path):
    with open(file_path) as ff:
        result = json.load(ff)
    return result


def read_csv_or_json(file_path):
    logger.debug(f"read_csv_or_json {file_path = }")
    if not file_path or not os.path.exists(file_path):
        return None
    if file_path.endswith(".json"):
        return read_json(file_path)
    if file_path.endswith(".csv"):
        return read_csv(file_path)
    return None


def get_next_dict_item(dict_value):
    return dict([next(iter(dict_value.items()))])


def parse_benchmark_instance(instance_path):
    logger.debug("\nreq_to_data_map:")
    req_to_data_map = read_csv_or_json(get_latest_matching_file(instance_path, REQ_TO_DATA_MAP_PATTERN))
    logger.debug(f"req_to_data_map: {get_next_dict_item(req_to_data_map) if req_to_data_map else None}")

    logger.debug("\nresult_perf:")
    result_perf = read_csv_or_json(get_latest_matching_file(instance_path, RESULT_PERF_PATTERN))
    result_perf = {kk: dict(zip(PERF_METRICS, vv)) for kk, vv in result_perf.items()} if result_perf else {}
    logger.debug(f"result_perf: {get_next_dict_item(result_perf) if result_perf else None}")

    logger.debug("\nresult_common:")
    result_common = read_csv_or_json(get_latest_matching_file(instance_path, RESULT_COMMON_PATTERN))
    logger.debug(f"result_common: {result_common if result_common else None}")

    logger.debug("\nresults_per_request:")
    results_per_request = read_csv_or_json(get_latest_matching_file(instance_path, RESULTS_PER_REQUEST_PATTERN))
    logger.debug(f"results_per_request: {get_next_dict_item(results_per_request) if results_per_request else None}")

    return dict(
        req_to_data_map=req_to_data_map if req_to_data_map else {},
        result_perf=result_perf if result_perf else {},
        result_common=result_common if result_common else {},
        results_per_request=results_per_request if results_per_request else {},
    )


""" parse_mindie_server_config """


def parse_mindie_server_config(service_config_path):
    logger.debug("\nmindie_service_config:")
    if service_config_path.endswith(".json"):  # config.json directly
        mindie_service_path = os.path.dirname(os.path.dirname(service_config_path))
    else:  # mindie service path
        mindie_service_path = service_config_path
        service_config_path = os.path.join(service_config_path, "conf", "config.json")
    logger.info(f"mindie_service_path: {mindie_service_path}, service_config_path: {service_config_path}")
    mindie_service_config = read_csv_or_json(service_config_path)
    
    logger.debug(
        f"mindie_service_config: {get_next_dict_item(mindie_service_config) if mindie_service_config else None}"
    )

    if mindie_service_config:
        mindie_server_log_path = mindie_service_config.get("LogConfig", {}).get("logPath", "logs/mindie-server.log")
        mindie_server_log_path = os.path.join(mindie_service_path, mindie_server_log_path)
    else:
        mindie_server_log_path = None
    return mindie_service_config, mindie_server_log_path


""" analyze """


def analyze(mindie_service_config, benchmark_instance, mindie_server_log_path, params: ProfilingParameters):
    import msserviceprofiler.msservice_advisor.profiling_analyze
    from msserviceprofiler.msservice_advisor.profiling_analyze.register import REGISTRY, ANSWERS

    logger.info("")
    logger.info("<think>")
    for name, analyzer in REGISTRY.items():
        logger.info(name)
        analyzer(mindie_service_config, benchmark_instance, mindie_server_log_path, params)
    logger.info("</think>")

    logger.info("")
    logger.info("<answer>")
    for suggesion_type in SUGGESTION_TYPES:
        for name, items in ANSWERS.get(suggesion_type, dict()).items():
            for action, reason in items:
                logger.info(f"[{suggesion_type}] {name}")
                logger.info(f"[action] {action}")
                logger.info(f"[reason] {reason}")
                logger.info("")
    logger.info("</answer>")


""" check_positive_integer """


def check_positive_integer(value):
    try:
        value = int(value)
    except Exception as e:
        raise ValueError(f"'{value}' cannot convert to a positive integer.") from e
    
    if value < 0:
        raise ValueError(f"'{value}' is not a positive integer.")
    
    return value


def arg_parse(subparsers):
    mindie_service_path = os.getenv(MIES_INSTALL_PATH, MINDIE_SERVICE_DEFAULT_PATH)

    parser = subparsers.add_parser(
        "advisor", formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="advisor for performance"
    )
    parser.add_argument(
        "-i", "--instance_path", type=str, default="instance", help="benchamrk instance output directory"
    )
    parser.add_argument(
        "-s", "--service_config_path", type=str, default=mindie_service_path, help="service config json path"
    )
    parser.add_argument(
        "-t",
        "--target",
        type=str_ignore_case,
        default="ttft",
        choices=list(TARGETS_MAP.keys()),
        help="profiling key target",
    )
    parser.add_argument(
        "-m",
        "--target_metrics",
        type=lambda xx: PERF_METRICS_MAP.get(str_ignore_case(xx), None),
        default="average",
        choices=PERF_METRICS,
        help="profiling key target metrics",
    )
    parser.add_argument(
        "-in", "--input_token_num", type=check_positive_integer, default=0, help="input token number"
    )
    parser.add_argument(
        "-out", "--output_token_num", type=check_positive_integer, default=0, help="output token number"
    )
    parser.add_argument("-tp", "--tp", type=check_positive_integer, default=0, help="tp")
    parser.add_argument("--log_level", "-l", default="info", choices=LOG_LEVELS_LOWER, help="specify log level.")
    parser.set_defaults(func=main)


def main(args):
    profiling_params = ProfilingParameters.extract_from_args(args)
    set_log_level(args.log_level)
    benchmark_instance = parse_benchmark_instance(args.instance_path)
    mindie_service_config, mindie_server_log_path = parse_mindie_server_config(args.service_config_path)
    analyze(mindie_service_config, benchmark_instance, mindie_server_log_path, profiling_params)


# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from abc import ABC, abstractmethod


class InvalidParameterError(Exception):
    pass


class BaseConstraint(ABC):
    description = ""
    
    def __init__(self, *, description=None):
        self.description = self.description if description is None else description

    def __str__(self):
        return self.description
    
    def __and__(self, other):
        from .logic import AndConstraint
        from .helper import make_constraint 
        
        return AndConstraint(self, make_constraint(other))

    def __or__(self, other):
        from .logic import OrConstraint
        from .helper import make_constraint   
        
        return OrConstraint(self, make_constraint(other))
    
    def __invert__(self):
        from .logic import NotConstraint
        return NotConstraint(self)
    
    def __call__(self, val):
        return self.is_satisfied_by(val)

    @abstractmethod
    def is_satisfied_by(self, val):
        pass
    

class BasePathConstraint(BaseConstraint):
    def __init__(self, *, description=None):
        super().__init__(description=description)
        
    def __str__(self):
        return self.description
        
    def _get_path_stat(self, path: str):
        st = None
        # does not support file descriptor
        if not isinstance(path, (str, os.PathLike)):
            return st
        
        try:
            st = os.stat(path, follow_symlinks=False)
        except OSError:
            pass
        return st
